package at.jku.ssw.fp.airline;

import java.time.LocalDate;
import java.time.LocalDateTime;
import java.time.LocalTime;
import java.util.ArrayList;
import java.util.List;
import java.util.Random;
import java.util.concurrent.CompletableFuture;

/**
 * Class representing an airline in the flight examples.
 *
 * @author Herbert Praehofer
 *
 */
public class Airline implements Retailer {

  /**
   * Array with some airline objects.
   */
  public static final Airline[] AIRLINES = new Airline[] {
      new Airline("LH", "www.lufthansa.com"),
      new Airline("AF", "www.airfrance.fr"), new Airline("BA", "www.britishairways.com"),
      new Airline("UA", "www.united.com"), new Airline("AA", "www.aa.com"), new Airline("AEY", "www.m.aeroflot.ru"),
      new Airline("EK", "www.emirates.com"), new Airline("RYR", "www.ryanair.com"),
      new Airline("SQ", "www.singaporeair.com") };

  /** Constant for the delay of retrieving a flight from an airline. */
  private static final int DELAY_CONST = 100;

  /** Constant for the variation of the delay of retrieving a flight from an airline. */
  private static final int DELAY_VARIATION = 2000;

  /** A random number generator */
  static final Random RAND = new Random();

  /** The code of the airline */
  public final String code;

  /** The name of the airline */
  public final String name;

  /**
   * Constructor setting code and name of the airline.
   *
   * @param code the code of the airline
   * @param name the name of the airline
   */
  public Airline(String code, String name) {
    super();
    this.code = code;
    this.name = name;
  }

  /**
   * Returns a flight provided by the airline.
   *
   * @param from the origin
   * @param to the destination
   * @param date the date
   * @return the flight object provided by the airline
   */
  public Flight findFlight(String from,
      String to,
      LocalDate date) {
    try {
      Thread.sleep(DELAY_CONST + RAND.nextInt(DELAY_VARIATION));
    } catch (InterruptedException e) {
    }
    return new Flight(from, to, this, code + String.valueOf((int) (Math.random() * 5000)),
        LocalDateTime.of(date, LocalTime.of(0, 0).plusHours(RAND.nextInt(24)).plusMinutes(RAND.nextInt(60))),
        200 + RAND.nextInt(200));
  }

  /**
   * Returns a completable future for a flight provided by the airline.
   *
   * @param from the origin
   * @param to the destination
   * @param date the date
   * @return a completable future for flight object provided by the airline
   */
  public CompletableFuture<Flight> findFlightAsync(
          String from,
          String to,
          LocalDate date) {
    return CompletableFuture.supplyAsync(() -> findFlight(from, to, date));
  }

  /**
   * Returns a list of flights offered by the airline.
   * @param from the origin
   * @param to the destination
   * @param date the date
   * @return the list of flights offered by the airline
   */
  public List<Flight> findFlights(String from, String to, LocalDate date) {
    List<Flight> flights = new ArrayList<Flight>();
    flights.add(findFlight(from, to, date));
    while (RAND.nextBoolean()) {
      flights.add(findFlight(from, to, date));
    }
    return flights;
  }

  /**
   * Returns a offer for a flight.
   *
   * @param flight the flight
   * @return the offer for the flight
   */
  public Offer getOffer(Flight flight) {
    return new Offer(this, flight, 200 + RAND.nextInt(120));
  }

  /**
   * Returns a string representation of the airline.
   *
   * @return the string representation
   */
  @Override
  public String toString() {
    return "Airline [code=" + code + ", name=" + name + "]";
  }

}
