package at.jku.ssw.fp.coll;

import java.io.Serializable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.function.BiFunction;
import java.util.function.BinaryOperator;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.stream.Stream;
import java.util.stream.StreamSupport;

/**
 * Generic functional list implementation. Functional lists are immutable.
 * Elements can be added in front of a functional list and
 * a new list with the new element as head and this list as tail is returned.
 *
 * @param <E> The type of elements
 *
 * @author Herbert Praehofer
 */
@SuppressWarnings("serial")
public abstract class FList<E> implements Iterable<E>, Serializable {

  /** The single object representing the empty list */
  @SuppressWarnings("rawtypes")
  private static FList NIL = new Nil();

  /**
   * Returns the empty list.
   *
   * @param <E> the element type
   * @return the empty list
   */
  @SuppressWarnings("unchecked")
  public static <E> FList<E> empty() {
    return (FList<E>) NIL;
  }

  /**
   * Creates a functional list with the given elements.
   *
   * @param <E> the element type
   * @param elems the array of elements for this list
   * @return the functional list with the given elements
   */
  @SuppressWarnings("unchecked")
  public static <E> FList<E> of(E... elems) {
    return Builder.of(elems).build();
  }

  /**
   * Creates a new list with the given element as head and this list as tail.
   *
   * @param elem the first element of the created list
   * @return the functional list with the given element as head and this list as tail
   */
  public FList<E> add(E elem) {
    return new Cons<E>(elem, this);
  }

  /**
   * Tests if this list is the empty list.
   *
   * @return {@code true}, if this list is the empty list; {@code false} otherwise
   */
  public abstract boolean isEmpty();

  /**
   * Returns the size of this list.
   *
   * @return the size of this list
   */
  public abstract int size();

  /**
   * Returns the head element of this list.
   *
   * @return the head element
   */
  public abstract E head();

  /**
   * Returns the tail list of this list.
   * @return the tail list
   */
  public abstract FList<E> tail();


  /**
   * Tests if the provided object is contained in this list.
   *
   * @param o the object to look for
   * @return {@code true} if object found,  {@code false} otherwise
   */
  public boolean contains(Object o) {
    for (E e : this) {
      if (e.equals(o))
        return true;
    }
    return false;
  }

  /**
   * Gets the element at index {@code idx}.
   * Throws an {@link java.lang.IndexOutOfBoundsException}
   * if the index is out of range.
   *
   * @param idx the position of the element
   * @return the element at position {@code idx}
   */
  public E get(int idx) {
    if (idx >= size()) {
      throw new IndexOutOfBoundsException("Index out of bounds");
    }
    int i = 0;
    for (E e : this) {
      if (i == idx)
        return e;
      i++;
    }
    return null; // should not happen
  }

  /**
   * Creates a new functional list with same elements as this list
   * but without the given object.
   *
   * @param obj the object to remove
   * @return a new functional list without the given object
   */
  public FList<E> remove(Object obj) {
    return removeRec(this, obj);
  }

  /**
   * Creates a new functional list with same elements as this list
   * but with the element at index  {@code idx} removed.
   * @param idx the index of the element to remove
   * @return a new functional list without the element at index {@code idx}
   */
  public FList<E> removeIdx(int idx) {
    FList<E> l = this;
    Builder<E> builder = Builder.create();
    for (int i = 0; i < idx; i++) {
      builder.add(l.head());
      l = l.tail();
    }
    if (idx == size() - 1) {
      return builder.build();
    } else {
      return builder.concat(l.tail());
    }
  }

  /**
   * Creates a new functional list with same elements as this list
   * but with the element at position {@code idx} replaced by the given element.
   *
   * @param idx the index of the element to replace
   * @param elem the new element
   * @return a new functional list without the element at index {@code idx} replaced
   *         by the new element
   */
  public FList<E> replaceIdx(int idx, E elem) {
    int i = 0;
    Builder<E> b = Builder.create();
    for (E e : this) {
      if (i == idx) {
        b.add(elem);
      } else {
        b.add(e);
      }
      i++;
    }
    return b.build();
  }

  /**
   * Returns an iterator for this list.
   *
   * @return the iterator for this list
   */
  public Iterator<E> iterator() {
    return new Iterator<E>() {
      FList<E> current = FList.this;

      @Override
      public boolean hasNext() {
        return !current.isEmpty();
      }

      @Override
      public E next() {
        E next = current.head();
        current = current.tail();
        return next;
      }
    };
  }

  /**
   * Creates a new functional list with the elements reversed.
   *
   * @return a new functional list with the elements reversed
   */
  public FList<E> reverse() {
    FList<E> l = FList.empty();
    for (E e : this) {
      l = l.add(e);
    }
    return l;
  }

  /**
   * Creates an object array with the elements of this list.
   *
   * @return an object array with the elements of this list
   */
  public Object[] toArray() {
    Object[] result = new Object[size()];
    int i = 0;
    for (E x : this)
      result[i++] = x;
    return result;
  }

  /**
   * Creates an array with the elements of this list.
   * The runtime type of the returned array is that
   * of the given array.
   *
   * @param a the array with the same type as the returned array
   * @return the array with the elements of this list
   */
  @SuppressWarnings("unchecked")
  public E[] toArray(E[] a) {
    if (a.length < size())
      a = (E[]) java.lang.reflect.Array.newInstance(a.getClass().getComponentType(), size());
    int i = 0;
    Object[] result = a;
    for (E x : this)
      result[i++] = x;

    if (a.length > size())
      a[size()] = null;

    return a;
  }

  /**
   * Returns a string representation of this list.
   *
   * @return a string representation of this list
   */
  @Override
  public String toString() {
    if (isEmpty())
      return "[]";
    StringBuilder b = new StringBuilder("[");
    for (E e : this) {
      b.append(e.toString());
      b.append(", ");
    }
    b.delete(b.length() - 2, b.length());
    b.append("]");
    return b.toString();
  }

  /**
   * Returns a functional list consisting of the results of applying the given
   * function to the elements of this list.
   *
   * @param <R> The element type of the new list
   * @param mapper a function to apply to each element
   * @return the new list with the mapped elements
   */
  public <R> FList<R> map(Function<? super E, ? extends R> mapper) {
    Builder<R> builder = Builder.create();
    for (E e : this)
      builder.add(mapper.apply(e));
    return builder.build();
  }

  /**
   * Returns a functional list consisting of the elements of this list that match
   * the given predicate.
   *
   * @param predicate a predicate to apply to each element to determine if it
   *                  should be included
   * @return the new list with the elements filtered
   */
  public FList<E> filter(Predicate<? super E> predicate) {
    Builder<E> builder = Builder.create();
    for (E e : this)
      if (predicate.test(e))
        builder.add(e);
    return builder.build();
  }


  /**
   * Returns a functional list consisting of the results of replacing each element of
   * this list with the contents of a mapped list produced by applying
   * the provided mapping function to each element.
   *
   * @param <R> The element type of the new list
   * @param mapper a function to apply to each element which produces a list
   *               of new values
   * @return the new list with the mapped elements
   */
  public <R> FList<R> flatMap(
      Function<? super E, ? extends FList<? extends R>> mapper) {
     Builder<R> builder = Builder.create();
     for (E e : this)
       for (R r : mapper.apply(e))
       builder.add(r);
     return builder.build();
   }


  /**
   * Finds an element that fulfills the given predicate.
   * Returns the found element in an {@link java.util.Optional},
   * with {@link java.util.Optional#empty()} if no element found.
   *
   * @param pred the predicate to test
   * @return the found element in an {@link java.util.Optional},
   *         and with {@link java.util.Optional#empty()} if no element was found
   */
  public Optional<E> find(Predicate<? super E> pred) {
    for (E e : this) {
      if (pred.test(e))
        return Optional.of(e);
    }
    return Optional.empty();
  }

  /**
   * Performs an action for each element of this list.
   *
   * @param action an function to perform on the elements
   */
  public void forEach(Consumer<? super E> action) {
    for (E e : this)
      action.accept(e);
  }

  /**
   * Performs a reduction on the elements of this list starting with provided
   * identity value and using the provided accumulation function.
   *
   * @param <R> The type of the result
   * @param identity the identity value for the combiner function
   * @param acc the accumulation function
   * @return the result of the reduction
   */
  public <R> R reduce(R identity, BiFunction<? super R, ? super E, ? extends R> acc) {
    R result = identity;
    for (E e : this)
      result = acc.apply(result, e);
    return result;
  }

  /**
   * Performs a reduction on the elements of this list starting with provided
   * identity value and using the provided combiner operator.
   * The result if the reduction is of the same type as the type of the elements.
   * A binary operation is used for reduction.
   *
   * @param identity the identity value for the combiner function
   * @param combiner the combiner operator
   * @return the result of the reduction
   */
  public E reduce(E identity, BinaryOperator<E> combiner) {
    E result = identity;
    for (E e : this)
      result = combiner.apply(result, e);
    return result;
  }

  /**
   * Creates a functional list with {@code n} elements generated
   * by the provided supplier function.
   *
   * @param <T> the type of the elements
   * @param n the number of elements
   * @param supplier the supplier function for creating the elements
   * @return the functional list with the generated elements
   */
  public static <T> FList<T> generate(int n, Supplier<T> supplier) {
    Builder<T> builder = Builder.create();
    for (int i = 0; i < n; i++)
      builder.add(supplier.get());
    return builder.build();
  }

  /**
   * Creates a sequential stream from this functional list.
   *
   * @return the stream object
   * @see java.util.stream.Stream
   */
  public Stream<E> stream() {
    return StreamSupport.stream(this.spliterator(),  false);
  }

  /**
   * Creates a parallel stream from this functional list.
   *
   * @return the stream object
   * @see java.util.stream.Stream
   */
  public Stream<E> parallelStream() {
    return StreamSupport.stream(this.spliterator(),  true);
  }

  /**
   * Helper function for remove elements.
   *
   * @param fList the list to remove the element
   * @param obj the object to remove
   * @return the functional list with the element removed
   */
  private FList<E> removeRec(FList<E> fList, Object obj) {
    LinkedList<E> ll = new LinkedList<E>();
    FList<E> fL = fList;
    while (!fL.isEmpty() && !fL.head().equals(obj)) {
      ll.addFirst(fL.head());
      fL = fL.tail();
    }
    if (fL.isEmpty())
      return fList;
    fL = removeRec(fL.tail(), obj);
    for (E e : ll)
      fL = fL.add(e);
    return fL;
  }


  /**
   * Class for the empty list.
   *
   * @param <E> the element type (used in type inference)
   */
  public static final class Nil<E> extends FList<E> {

    /**
     * Private constructor.
     */
    private Nil() {
    }

    /**
     * Returns {@code true} as this represents the empty list.
     *
     * @return {@code true}
     */
    @Override
    public boolean isEmpty() {
      return true;
    }

    /**
     * Returns {@code 0} as size of the empty list.
     *
     * @return {@code 0}
     */
    @Override
    public int size() {
      return 0;
    }

    /**
     * Throws a {@link NoSuchElementException}.
     *
     * @return nothing
     */
    @Override
    public E head() {
      throw new NoSuchElementException("No head of empty list");
    }

    /**
     * Throws a {@link NoSuchElementException}.
     *
     * @return nothing
     */
   @Override
    public FList<E> tail() {
      throw new NoSuchElementException("No tail of empty list");
    }

  }

  /**
   * Class representing a functional list with a head element and tail list.
   *
   * @param <E> the element type
   */
  public static final class Cons<E> extends FList<E> {

    /** The head element */
    private final E head;

    /** The tail list */
    private final FList<E> tail;

    /** The number of elements */
    private final int size;

    /**
     * Private constructor initializing head and tail.
     *
     * @param head the head element
     * @param tail the tail list
     */
    private Cons(E head, FList<E> tail) {
      this.head = head;
      this.tail = tail;
      this.size = tail.size() + 1;
    }

    /**
     * Returns {@code false} as this list is not empty.
     *
     * @return {@code false}
     */
    @Override
    public boolean isEmpty() {
      return false;
    }

    /**
     * Returns the number of elements of this functional list.
     *
     * @return the number of elements in this list.
     */
    @Override
    public int size() {
      return size;
    }

    /**
     * Returns the head element of this list.
     *
     * @return the head element
     */
    @Override
    public E head() {
      return head;
    }

    /**
     * Returns the tail list of this list.
     *
     * @return the tail list
     */
    @Override
    public FList<E> tail() {
      return tail;
    }

    /**
     * Returns a hash code value for the list. It is computed from the
     * hash code of the head element and the hash code of the tail list.
     *
     * @return the hash code
     */
    @Override
    public int hashCode() {
      final int prime = 31;
      int result = 1;
      result = prime * result + size;
      result = prime * result + ((tail == null) ? 0 : tail.hashCode());
      result = prime * result + ((head == null) ? 0 : head.hashCode());
      return result;
    }

    /**
     * Indicates whether some other object is "equal to" this list.
     * The object is equal to this list, if it is a functional list
     * and the head element and the tail list are equal.
     *
     * @return  {@code true} if this list is the same as the argument;
     *          {@code false} otherwise.
     */
    @Override
    public boolean equals(Object obj) {
      if (this == obj)
        return true;
      if (obj == null)
        return false;
      if (getClass() != obj.getClass())
        return false;
      @SuppressWarnings("rawtypes")
      Cons other = (Cons) obj;
      if (size != other.size)
        return false;
      if (head == null) {
        if (other.head != null)
          return false;
      } else if (!head.equals(other.head))
        return false;
      return tail.equals(other.tail);
    }

  }

  /**
   * Mutable builder for building functional lists.
   * It internally uses a mutable linked list for collecting
   * the elements. Then allows to create the immutable functional
   * list with the collected elements.
   *
   * @param <E> the type of the elements
   */
  public static class Builder<E> {

    /**
     * Creates a builder with the provided elements
     * already collected.
     *
     * @param <E> the element type
     * @param es the elements to collect
     * @return the builder with the elements collected
     */
    public static <E> Builder<E> of(Iterable<? extends E> es) {
      return new Builder<E>(es);
    }

    /**
     * Creates a builder with the provided elements
     * already collected.
     *
     * @param <E> the element type
     * @param es the elements to collect
     * @return the builder with the elements collected
     */
    @SuppressWarnings("unchecked")
    public static <E> Builder<E> of(E... es) {
      return new Builder<E>(es);
    }

    /**
     * Creates a new builder for collecting elements and
     * building a functional list.
     *
     * @param <E> the element type
     * @return a new builder
     */
    public static <E> Builder<E> create() {
      return new Builder<E>();
    }

    /** The mutable linked list for collecting the elements. */
    private final LinkedList<E> list;

    /**
     * Private constructor for creating an empty builder.
     */
    private Builder() {
      list = new LinkedList<E>();
    }

    /**
     * Private constructor for creating a builder with the
     * provided elements already collected.
     *
     * @param es the elements to collect
     */
    private Builder(E[] es) {
      list = new LinkedList<E>();
      for (E e : es)
        add(e);
    }

    /**
     * Private constructor for creating a builder where the
     * elements in the provided iterable get collected.
     *
     * @param es the elements to collect
     */
    private Builder(Iterable<? extends E> es) {
      list = new LinkedList<E>();
      for (E e : es)
        add(e);
    }

    /**
     * Adds an element to the builder.
     *
     * @param e the element to add
     */
    public void add(E e) {
      list.addFirst(e);
    }

    /**
     * Adds all elements in the iterable to this builder.
     *
     * @param es the elements to add
     */
    public void addAll(Iterable<? extends E> es) {
      for (E e : es)
        add(e);
    }

    /**
     * Adds all elements in the array to this builder.
     *
     * @param es the elements to add
     */
   public void addAll(E[] es) {
      for (E e : es)
        add(e);
    }

   /**
    * Adds all elements in the given list to this builder.
    *
    * @param fList the elements to add
    * @return the concatenated list
    */
   public FList<E> concat(FList<E> fList) {
      for (E e : list) {
        fList = fList.add(e);
      }
      return fList;
    }

   /**
    * Creates a functional list with the elements collected in the builder.
    *
    * @return the functional list with the colleced elements
    */
    public FList<E> build() {
      FList<E> fList = FList.empty();
      for (E e : list) {
        fList = fList.add(e);
      }
      return fList;
    }

  }

}
