package at.jku.ssw.fp.coll;

import java.io.Serializable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Set;
import java.util.function.Consumer;
import java.util.function.Function;
import java.util.function.Predicate;

/**
 * Generic functional set implementation. Functional sets are immutable.
 *
 * This implementation largely is a copy of the function list {@link FList}
 * class with the difference that equal elements are stored only once.
 * Moreover, only a subset of methods are supported.
 *
 * @param <E> The type of elements
 *
 * @author Herbert Praehofer
 */
@SuppressWarnings("serial")
public abstract class FSet<E> implements Iterable<E>, Serializable {

  /** The single object representing the empty set */
  @SuppressWarnings("rawtypes")
  private static final FSet EMPTY = new Empty();

  /**
   * Returns the empty set.
   *
   * @param <E> the element type
   * @return the empty list
   */
  @SuppressWarnings("unchecked")
  public static <E> FSet<E> empty() {
    return (FSet<E>) EMPTY;
  }

  /**
   * Creates a functional set with the given elements.
   *
   * @param <E>   the element type
   * @param elems the array of elements for this set
   * @return the functional set with the given elements
   */
  @SuppressWarnings("unchecked")
  public static <E> FSet<E> of(E... elems) {
    return Builder.of(elems).build();
  }

  /**
   * Creates a functional set from a list of elements.
   *
   * @param <E>   the element type
   * @param elems the list of elements
   * @return the functional set with the given elements
   */
  public static <E> FSet<E> of(List<E> elems) {
    return Builder.of(elems).build();
  }

  /**
   * Creates a functional set from a set of elements.
   *
   * @param <E>   the element type
   * @param elems the set of elements
   * @return the functional set with the given elements
   */
  public static <E> FSet<E> of(Set<E> elems) {
    return Builder.of(elems).build();
  }

  /**
   * Creates a new set with the elements of this set and the new element.
   *
   * @param elem the new element
   * @return the functional set with the elements of this set and the new element
   */
  public FSet<E> add(E elem) {
    if (!this.contains(elem)) {
      return new Cons<E>(elem, this);
    } else {
      return this;
    }
  }

  /**
   * Tests if this list is the empty set.
   *
   * @return {@code true}, if this set is the empty set; {@code false} otherwise
   */
  public abstract boolean isEmpty();

  /**
   * Returns the size of this set.
   *
   * @return the size of this set
   */
  public abstract int size();

  /**
   * Returns the first element of this set.
   *
   * @return the first element
   */
  public abstract E head();

  /**
   * Returns the tail set for this set.
   *
   * @return the tail set
   */
  public abstract FSet<E> tail();

  /**
   * Tests if the provided object is contained in this set.
   *
   * @param o the object to look for
   * @return {@code true} if object found, {@code false} otherwise
   */
  public boolean contains(Object o) {
    for (E e : this) {
      if (e.equals(o))
        return true;
    }
    return false;
  }

  /**
   * Returns an iterator for this set.
   *
   * @return the iterator for this set
   */
  public Iterator<E> iterator() {
    return new Iterator<E>() {
      FSet<E> current = FSet.this;

      @Override
      public boolean hasNext() {
        return !current.isEmpty();
      }

      @Override
      public E next() {
        E next = current.head();
        current = current.tail();
        return next;
      }
    };
  }

  /**
   * Returns a string representation of this set.
   *
   * @return a string representation of this set
   */
  @Override
  public String toString() {
    if (isEmpty())
      return "{}";
    StringBuilder b = new StringBuilder("{");
    for (E e : this) {
      b.append(e.toString());
      b.append(", ");
    }
    b.delete(b.length() - 2, b.length());
    b.append("}");
    return b.toString();
  }

  /**
   * Returns a functional set consisting of the results of applying the given
   * function to the elements of this set.
   *
   * @param <R>    The element type of the new set
   * @param mapper a function to apply to each element
   * @return the new set with the mapped elements
   */
  public <R> FSet<R> map(Function<? super E, ? extends R> mapper) {
    Builder<R> builder = Builder.create();
    for (E e : this) {
      builder.add(mapper.apply(e));
    }
    return builder.build();
  }

  /**
   * Returns a functional set consisting of the elements of this set that match
   * the given predicate.
   *
   * @param predicate a predicate to apply to each element to determine if it
   *                  should be included
   * @return the new set with the elements filtered
   */
  public FSet<E> filter(Predicate<? super E> predicate) {
    Builder<E> builder = Builder.create();
    for (E e : this) {
      if (predicate.test(e))
        builder.add(e);
    }
    return builder.build();
  }

  /**
   * Returns a functional set consisting of the results of replacing each element
   * of this set with the contents of a mapped set produced by applying the
   * provided mapping function to each element.
   *
   * @param <R>    The element type of the new set
   * @param mapper a function to apply to each element which produces a set of new
   *               values
   * @return the new set with the mapped elements
   */
  public <R> FSet<R> flatMap(Function<? super E, ? extends FSet<R>> mapper) {
    Builder<R> builder = Builder.create();
    for (E e : this) {
      for (R r : mapper.apply(e)) {
        builder.add(r);
      }
    }
    return builder.build();
  }

  /**
   * Performs an action for each element of this set.
   *
   * @param action an function to perform on the elements
   */
  public void forEach(Consumer<? super E> action) {
    for (E e : this)
      action.accept(e);
  }

  /**
   * Creates an object array with the elements of this set.
   *
   * @return an object array with the elements of this set
   */
  public Object[] toArray() {
    Object[] result = new Object[size()];
    int i = 0;
    for (E x : this)
      result[i++] = x;
    return result;
  }

  /**
   * Creates an array with the elements of this set. The the runtime type of the
   * returned array is that of the given array.
   *
   * @param a the array with the same type as the returned array
   * @return the array with the elements of this set
   */
  @SuppressWarnings("unchecked")
  public E[] toArray(E[] a) {
    if (a.length < size())
      a = (E[]) java.lang.reflect.Array.newInstance(a.getClass().getComponentType(), size());
    int i = 0;
    Object[] result = a;
    for (E x : this)
      result[i++] = x;

    if (a.length > size())
      a[size()] = null;

    return a;
  }

  // inner classes

  /**
   * Class for the empty set.
   *
   * @param <E> the element type (used in type inference)
   */
  public static final class Empty<E> extends FSet<E> {

    /**
     * Constructor.
     */
    private Empty() {
    }

    /**
     * Returns {@code true} as this represents the empty set.
     *
     * @return {@code true}
     */
    @Override
    public boolean isEmpty() {
      return true;
    }

    /**
     * Returns {@code 0} as size of the empty set.
     *
     * @return {@code 0}
     */
    @Override
    public int size() {
      return 0;
    }

    /**
     * Throws a {@link NoSuchElementException}.
     *
     * @return nothing
     */
    @Override
    public E head() {
      throw new NoSuchElementException("No head of empty list");
    }

    /**
     * Throws a {@link NoSuchElementException}.
     *
     * @return nothing
     */
    @Override
    public FSet<E> tail() {
      throw new NoSuchElementException("No tail of empty list");
    }

  }

  /**
   * Class representing a functional set.
   *
   * @param <E> the element type
   */
  public static final class Cons<E> extends FSet<E> {

    /** The head element */
    private final E head;

    /** The tail list */
    private final FSet<E> tail;

    /** The number of elements */
    private final int size;

    /**
     * Private constructor initializing head and tail.
     *
     * @param head the head element
     * @param tail the tail set
     */
    private Cons(E head, FSet<E> tail) {
      this.head = head;
      this.tail = tail;
      this.size = tail.size() + 1;
    }

    /**
     * Returns {@code false} as this list is not empty.
     *
     * @return {@code false}
     */
    @Override
    public boolean isEmpty() {
      return false;
    }

    /**
     * Returns the number of elements of this functional set.
     *
     * @return the number of elements in this set.
     */
    @Override
    public int size() {
      return size;
    }

    /**
     * Returns the first element of this set.
     *
     * @return the first element
     */

    @Override
    public E head() {
      return head;
    }

    /**
     * Returns the tail of this set.
     *
     * @return the tail
     */
    @Override
    public FSet<E> tail() {
      return tail;
    }

    /**
     * Returns a hash code value for the set. It is computed from the
     * hash code of the head element and the hash code of the tail.
     *
     * @return the hash code
     */
    @Override
    public int hashCode() {
      final int prime = 31;
      int result = 1;
      result = prime * result + size;
      result = prime * result + ((tail == null) ? 0 : tail.hashCode());
      result = prime * result + ((head == null) ? 0 : head.hashCode());
      return result;
    }

    /**
     * Indicates whether some other object is "equal to" this set.
     * The object is equal to this set, if it is a functional set
     * and the head element and the tail are equal.
     *
     * @return  {@code true} if this set is the same as the argument;
     *          {@code false} otherwise.
     */
    @Override
    public boolean equals(Object obj) {
      if (this == obj)
        return true;
      if (obj == null)
        return false;
      if (getClass() != obj.getClass())
        return false;
      @SuppressWarnings("rawtypes")
      Cons other = (Cons) obj;
      if (size != other.size)
        return false;
      if (head == null) {
        if (other.head != null)
          return false;
      } else if (!head.equals(other.head))
        return false;
      return tail.equals(other.tail);
    }
  }

  /**
   * Mutable builder for building functional sets.
   * It internally uses a mutable linked list for collecting
   * the elements. Then allows to create the immutable functional
   * list with the collected elements.
   *
   * @param <E> the type of the elements
   */
  public static class Builder<E> {

    /**
     * Creates a builder with the provided elements
     * already collected.
     *
     * @param <E> the element type
     * @param es the elements to collect
     * @return the builder with the elements collected
     */
    @SuppressWarnings("unchecked")
    public static <E> Builder<E> of(E... es) {
      return new Builder<E>(es);
    }

    /**
     * Creates a builder with the provided elements
     * already collected.
     *
     * @param <E> the element type
     * @param es the elements to collect
     * @return the builder with the elements collected
     */
    public static <E> Builder<E> of(Iterable<? extends E> es) {
      return new Builder<E>(es);
    }

    /**
     * Creates a new builder for collecting elements and
     * building a functional set.
     *
     * @param <E> the element type
     * @return a new builder
     */
    public static <E> Builder<E> create() {
      return new Builder<E>();
    }

    /** The mutable linked list for collecting the elements. */
    private final LinkedList<E> list;

    /**
     * Private constructor for creating an empty builder.
     */
    private Builder() {
      list = new LinkedList<E>();
    }

    /**
     * Private constructor for creating a builder with the
     * provided elements already collected.
     *
     * @param es the elements to collect
     */
    private Builder(E[] es) {
      list = new LinkedList<E>();
      for (E e : es)
        add(e);
    }

    /**
     * Private constructor for creating a builder where the
     * elements in the provided iterable get collected.
     *
     * @param es the elements to collect
     */
    private Builder(Iterable<? extends E> es) {
      list = new LinkedList<E>();
      for (E e : es)
        add(e);
    }

    /**
     * Adds an element to the builder, if not already contained.
     *
     * @param e the element to add
     */
    public void add(E e) {
      if (!list.contains(e))
        list.addFirst(e);
    }

    /**
     * Creates an functional set from the collected elements.
     *
     * @return the functional set with the collected elements
     */
    public FSet<E> build() {
      FSet<E> fSet = FSet.empty();
      for (E e : list) {
        fSet = fSet.add(e);
      }
      return fSet;
    }

  }

}
