package at.jku.ssw.fp.coll;

import java.io.Serializable;
import java.util.NoSuchElementException;

/**
 * Binary tree with elements which are generic. Trees are immutable.
 * This structure closely corresponds to the functional list structure
 * {@link FList}
 *
 * @param <E> The type of elements
 *
 * @author Herbert Praehofer
 */
@SuppressWarnings("serial")
public abstract class Tree<E> implements Serializable {

  /** The single object representing the empty tree */
  @SuppressWarnings("rawtypes")
  private static final Tree EMPTY = new Empty();

  /**
   * Returns the empty tree.
   *
   * @param <E> the element type
   * @return the empty tree
   */
  @SuppressWarnings("unchecked")
  public static <E> Tree<E> empty() {
    return (Tree<E>) EMPTY;
  }

  /**
   * Creates a functional tree node with a value and a left and right subtree.
   *
   * @param <E> the element type
   * @param value the value of this tree node
   * @param left the left subtree
   * @param right the right subtree
   * @return the tree node with value and left and right subtrees
   */
  public static <E> Tree<E> of(E value, Tree<E> left, Tree<E> right) {
    return new Node<E>(value, left, right);
  }


  /**
   * Creates a functional tree node with a value and empty left and right subtrees.
   *
   * @param <E> the element type
   * @param value the value of this tree node
   * @return the tree node with value and empty left and right subtrees
   */
  public static <E> Tree<E> of(E value) {
    return new Node<E>(value, empty(), empty());
  }

  /**
   * Returns the value of this tree node.
   *
   * @return the value of this tree node
   */
  public abstract E value();

  /**
   * Returns the left subtree of this tree node.
   *
   * @return the left subtree of this tree node
   */
  public abstract Tree<E> left();

  /**
   * Returns the right subtree of this tree node.
   *
   * @return the right subtree of this tree node
   */
  public abstract Tree<E> right();

  /**
   * Tests if this tree is the empty tree.
   *
   * @return {@code true}, if this tree is the empty tree; {@code false} otherwise
   */
  public abstract boolean isEmpty();

  /**
   * Tests if the left subtree is not empty.
   *
   * @return {@code true} if the left subtree is not empty; {@code false} otherwise
   */
  public abstract boolean leftExists();

  /**
   * Tests if the right subtree is not empty.
   *
   * @return {@code true} if the right subtree is not empty; {@code false} otherwise
   */
  public abstract boolean rightExists();

  /**
   * Class for the empty tree.
   *
   * @param <E> the element type (used in type inference)
   */
  public static class Empty<E> extends Tree<E> {

    /**
     * Constructor
     */
    private Empty() {}


    /**
     * Throws a {@link NoSuchElementException}.
     *
     * @return nothing
     */
    @Override
    public E value() {
      throw new NoSuchElementException("No value of empty tree");
    }


    /**
     * Throws a {@link NoSuchElementException}.
     *
     * @return nothing
     */
    @Override
    public Tree<E> left() {
      throw new NoSuchElementException("No left subtree of empty tree");
    }


    /**
     * Throws a {@link NoSuchElementException}.
     *
     * @return nothing
     */
    @Override
    public Tree<E> right() {
      throw new NoSuchElementException("No right subtree of empty tree");
    }

    /**
     * Returns {@code true} as this represents the empty tree.
     *
     * @return {@code true}
     */
    @Override
    public boolean isEmpty() {
      return true;
    }

    /**
     * Returns {@code false} as the empty tree does not have a left subtree.
     *
     * @return {@code false}
     */
    @Override
    public boolean leftExists() {
      return false;
    }

    /**
     * Returns {@code false} as the empty tree does not have a right subtree.
     *
     * @return {@code false}
     */
    @Override
    public boolean rightExists() {
      return false;
    }

    /**
     * Returns an empty string as string representation of the empty tree.
     *
     * @return the empty string
     */
    @Override
    public String toString() {
      return "";
    }

  }

  /**
   * Class representing a tree node with a value and a left and right subtree.
   *
   * @param <E> the element type
   */
  public static class Node<E> extends Tree<E> {

    /** The value of this tree node. */
    private final E value;

    /** The left subtree of this tree node; possibly empty. */
    private final Tree<E> left;

    /** The right subtree of this tree node; possibly empty. */
    private final Tree<E> right;

    /**
     * Private constructor initializing value and left and right subtrees.
     *
     * @param value the value of this tree node
     * @param left the left subtree of this tree node
     * @param right the right subtree of this tree node
     */
    private Node(E value, Tree<E> left, Tree<E> right) {
      super();
      this.value = value;
      this.left = left;
      this.right = right;
    }

    /**
     * Returns the value of this tree node.
     *
     * @return the value of this node
     */
    @Override
    public E value() {
      return value;
    }

    /**
     * Returns the left subtree of this tree node
     *
     * @return the left subtree of this tree node
     */
    @Override
    public Tree<E> left() {
      return left;
    }

    /**
     * Returns the right subtree of this tree node
     *
     * @return the right subtree of this tree node
     */
    @Override
    public Tree<E> right() {
      return right;
    }

    /**
     * Returns {@code false} as this tree node is not the empty tree node.
     *
     * @return {@code false}
     */
    @Override
    public boolean isEmpty() {
      return false;
    }

    /**
     * Test if the left subtree is not empty.
     *
     * @return {@code true} if the the left subtree is not empty;
     *         {@code false} otherwise
     */
    @Override
    public boolean leftExists() {
      return !left.isEmpty();
    }

    /**
     * Test if the right subtree is not empty.
     *
     * @return {@code true} if the the right subtree is not empty;
     *         {@code false} otherwise
     */    @Override
    public boolean rightExists() {
      return !right.isEmpty();
    }

   /**
    * Returns a string representation of this tree node.
    *
    * @return a string representation of this tree node
    */
    @Override
    public String toString() {
      return "(" + (leftExists() ? (left.toString() + "<-") : "") + value.toString()
          + (rightExists() ? ("->" + right.toString()) : "") + ")";
    }

    /**
     * Returns a hash code value for the tree node. It is computed from the
     * hash code of the value and the left and right substrees.
     *
     * @return the hash code
     */
   @Override
    public int hashCode() {
      final int prime = 31;
      int result = 1;
      result = prime * result + ((left == null) ? 0 : left.hashCode());
      result = prime * result + ((right == null) ? 0 : right.hashCode());
      result = prime * result + ((value == null) ? 0 : value.hashCode());
      return result;
    }

    /**
     * Indicates whether some other object is "equal to" this tree.
     * The object is equal to this tree, if it is a tree node
     * and the value and the left and right subtrees are equal.
     *
     * @return  {@code true} if this tree is the same as the argument;
     *          {@code false} otherwise.
     */
    @Override
    public boolean equals(Object obj) {
      if (this == obj)
        return true;
      if (obj == null)
        return false;
      if (getClass() != obj.getClass())
        return false;
      @SuppressWarnings("rawtypes")
      Node other = (Node) obj;
      if (left == null) {
        if (other.left != null)
          return false;
      } else if (!left.equals(other.left))
        return false;
      if (right == null) {
        if (other.right != null)
          return false;
      } else if (!right.equals(other.right))
        return false;
      if (value == null) {
        if (other.value != null)
          return false;
      } else if (!value.equals(other.value))
        return false;
      return true;
    }

  }

}
