package at.jku.ssw.fp.sect03_3;

import java.util.Optional;

import at.jku.ssw.fp.coll.FList;
import at.jku.ssw.fp.sect04_3.puzzle.Pos;

/**
 * Demo showing difference of solutions to NQueens problem using mutable
 * and immutable data structures.
 *
 * @author Herbert Praehofer
 */
public class Demo2_NQueens {

  private static final int N = 8;

  public static void main(String[] args) {
    boolean[][] board = new boolean[N][N];
    boolean solved = solveNQueens(board, 0);

    System.out.println(solved);
    print(board);

    Optional<FList<Pos>> solution = solveNQueens(FList.empty(), 0);

    System.out.println();
    print(solution.get());

  }

  /**
   * Solves N-Queens problem with mutable array.
   *
   * @param board mutable array representing the board
   * @param col next column
   * @return {@code true} if successful
   */
  static boolean solveNQueens(boolean[][] board, int col) {
    if (col >= board.length) {
      return true;
    } else {
      for (int i = 0; i < board.length; i++) {
        if (isSafe(board, i, col)) {
          board[i][col] = true;
          if (solveNQueens(board, col + 1))
            return true;
          board[i][col] = false; // Backtracking
        }
      }
      return false;
    }
  }

  /**
   * Solves N-Queens problem using immutable lists
   *
   * @param board the list of positions representing current board
   * @param col next column
   * @return optional with list of positions representing solution
   *           with empty optional if no solution found
   */
  static Optional<FList<Pos>> solveNQueens(FList<Pos> board, int col) {
    if (col >= N) {
      return Optional.of(board);
    } else {
      for (int i = 0; i < N; i++) {
        if (isSafe(board, i, col)) {
          Optional<FList<Pos>> optSol = solveNQueens(board.add(Pos.of(i, col)), col + 1);
          if (optSol.isPresent())
            return optSol;
        }
      }
      return Optional.empty();
    }
  }

  /**
   * Checks if position is safe for setting queen.
   *
   * @param board the mutable array representing the board
   * @param row the row index
   * @param col the column index
   * @return {@code true} if position is safe
   */
  private static boolean isSafe(boolean[][] board, int row, int col) {
    for (int i = 0; i < board.length; i++) {
      for (int j = 0; j < board[i].length; j++) {
        if (board[i][j] && (i == row || j == col || inDiag(i, j, row, col))) {
          return false;
        }
      }
    }
    return true;
  }

  /**
   * Checks if position is safe for setting queen.
   *
   * @param board the list of position currently set
   * @param row the row index
   * @param col the column index
   * @return {@code true} if position is safe
   */
  private static boolean isSafe(FList<Pos> board, int row, int col) {
    for (Pos pos : board) {
      if (pos.row == row || pos.col == col || inDiag(pos.row, pos.col, row, col)) {
        return false;
      }
    }
    return true;
  }

  /**
   * Checks if positions i/ j and row/col are in one diagonal.
   *
   * @param i the row coordinate of the first position
   * @param j the column coordinate of the first position
   * @param row the row coordinate of the second position
   * @param col the column coordinate of the second position
   * @return {@code true} if the positions are in one diagonal
   */
  private static boolean inDiag(int i, int j, int row, int col) {
    int dx = Math.abs(i - row);
    int dy = Math.abs(j - col);
    return dx == dy;
  }

  /**
   * Prints the board.
   *
   * @param board the board in form of a mutable array
   */
  private static void print(boolean[][] board) {
    for (int i = 0; i < N; i++) {
      System.out.println("\n-----------------");
      for (int j = 0; j < N; j++) {
        if (board[i][j]) {
          System.out.print("|X");
        } else {
          System.out.print("| ");
        }
      }
      System.out.print("| ");
    }
    System.out.println("\n-----------------");
  }

  /**
   * Prints the board.
   *
   * @param board the board in form of a immutable list
   */
  private static void print(FList<Pos> board) {
    for (int i = 0; i < N; i++) {
      System.out.println("\n-----------------");
      for (int j = 0; j < N; j++) {
        if (board.contains(Pos.of(i, j))) {
          System.out.print("|X");
        } else {
          System.out.print("| ");
        }
      }
      System.out.print("| ");
    }
    System.out.println("\n-----------------");
  }

}
