package at.jku.ssw.fp.sect04_2.fp;

import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.function.IntPredicate;

/**
 * Class representing the results of a student. 
 * 
 * @author Herbert Praehofer 
 */
public class Results {

  /** Number of assignments */
  private static final int N = 10;

  /** The student id */
  private final String id;
  
  /** The name of the student */
  private final String name;
  
  /** The array of scores for the different assignments */
  private final int[] scores;

  /**
   * Constructor initalizing id and name.
   * 
   * @param id the student id
   * @param name the student name
   */
  public Results(String id, String name) {
    this.id = id;
    this.name = name;
    this.scores = new int[N];
  }

  /**
   * Returns the id of the student. 
   * 
   * @return the id of the student
   */
  public String getStdId() {
    return id;
  }

  /**
   * Returns the name of the student. 
   * 
   * @return the name of the student
   */
  public String getName() {
    return name;
  }

  /**
   * Returns the score of an assignment.
   * 
   * @param assignment the number of the assignment. 
   * @return the score of the assignment
   */
  public int getScore(int assignment) {
    assert assignment >= 1 && assignment <= N;
    return scores[assignment - 1];
  }

  /**
   * Returns the array of the scores for the {@link N} assignments. 
   * 
   * @return the array of the scores
   */
  public int[] getScores() {
    return scores;
  }

  /**
   * Sets the score of an assignment. 
   * 
   * @param assignment the number of the assignment 
   * @param score the score for the assignment
   */
  public void setScore(int assignment, int score) {
    assert assignment >= 1 && assignment <= N;
    assert score >= 0 && assignment <= N;
    scores[assignment - 1] = score;
  }

  /** 
   * Sets the scores for the {@link N} assignments. 
   * 
   * @param scores the array with the scores
   */
  public void setScores(int... scores) {
    assert (scores.length == 10);
    for (int i = 1; i <= N; i++) {
      setScore(i, scores[i - 1]);
    }
  }

  /**
   * Computes the grades using a function object to compute the grades 
   * from the assignment scores. 
   * 
   * @param gradingFn the function object  
   * @return the grade
   */
  public Grade getGrade(Function<Results, Grade> gradingFn) {
    return gradingFn.apply(this);
  }

  /**
   * Counts the scores fulfilling a given predicate. 
   * 
   * @param countPredicate the predicate to test the scores
   * @return the number of scores fulfilling the given predicate
   */
  public int countScores(IntPredicate countPredicate) {
    int n = 0;
    for (int score : scores) {
      if (countPredicate.test(score)) {
        n++;
      }
    }
    return n;
  }

  /**
   * Computes a result from the scores. 
   * Uses an accumulator and a finalizer function.
   * 
   * @param <A> the type of the start and accumulated value
   * @param <R> the type of the result
   * @param start the start value
   * @param accumulator the accumulator function
   * @param finalizer the finalizer function
   * @return the computation result
   */
  public <A, R> R compute(A start, BiFunction<? super A, Integer, ? extends A> accumulator,
      Function<? super A, ? extends R> finalizer) {
    A a = start;
    for (int i = 1; i <= 10; i++) {
      a = accumulator.apply(a, this.getScore(i));
    }
    return finalizer.apply(a);
  }

}
