package at.jku.ssw.fp.sect04_2.fp;

import java.util.function.BiFunction;
import java.util.function.Function;
import java.util.function.IntPredicate;
import java.util.function.Predicate;

import at.jku.ssw.fp.coll.FList;

/**
 * Class for representing and maintaining a list of result records
 * for a set of students.
 *
 * @author Herbert Praehofer
 */
public class ResultsList {

  /** The list of results for a set of students */
  private FList<Results> resultsList;

  /** The function object to compute the grades */
  private Function<Results, Grade> gradingFn;

  /**
   * Constructor initializing the list of results with an empty list.
   */
  public ResultsList() {
    resultsList = FList.of();
  }

  /**
   * Adds a student to the list.
   * @param id the id of the student
   * @param name the name of the student
   */
  public void addStudent(String id, String name) {
    resultsList = resultsList.add(new Results(id, name));
  }

  /**
   * Sets the score for a student for an assignment.
   *
   * @param id the id of the student
   * @param assignment the assignment number
   * @param score the score
   */
  public void setScore(String id, int assignment, int score) {
    assert (assignment >= 1 && assignment <= 10);
    assert (score >= 0 && score <= 10);
    getResults(id).setScore(assignment, score);
  }

  /**
   * Sets the scores for a student.
   *
   * @param id the id of the student
   * @param scores the scores for the assignments
   */
  public void setScores(String id, int... scores) {
    assert (scores.length == 10);
    getResults(id).setScores(scores);
  }

  /**
   * Gets the scores of a student.
   *
   * @param id the id of the student
   * @return the array with the scores
   */
  public int[] getScores(String id) {
    return getResults(id).getScores();
  }

  /**
   * Gets the result object for a student.
   *
   * @param id the id of the student
   * @return the result object
   */
  public Results getResults(String id) {
    for (Results r : resultsList) {
      if (r.getStdId().equals(id))
        return r;
    }
    return null;
  }

  /**
   * Sets the function object to compute the grades.
   *
   * @param gradingFn the function object to compute the grades
   */
  public void setGradingFunction(Function<Results, Grade> gradingFn) {
    this.gradingFn = gradingFn;
  }

  /**
   * Gets the grade for a student computed from the assignment scores.
   * @param id the student id
   * @return the grade for the student.
   */
  public Grade getGrade(String id) {
    return getResults(id).getGrade(gradingFn);
  }

  /**
   * Counts the scores of a student fulfilling a given predicate.
   * @param id the student id
   * @param scorePred the predicate
   * @return the number of scores fulfilling the predicate
   */
  public int countScores(String id, IntPredicate scorePred) {
    int n = 0;
    for (int score : getScores(id)) {
      if (scorePred.test(score))
        n++;
    }
    return n;
  }

  /**
   * Computes a result from the scores for a student.
   * Uses an accumulator and a finalizer function.
   *
   * @param <A> the type of the start and accumulated value
   * @param <R> the type of the result
   * @param id the student id
   * @param start the start value
   * @param accumulator the accumulator function
   * @param finalizer the finalizer function
   * @return the computation result
   */
  public <A, R> R compute(String id, A start, BiFunction<? super A, Integer, ? extends A> accumulator,
      Function<? super A, ? extends R> finalizer) {
    return getResults(id).compute(start, accumulator, finalizer);
  }

  /**
   * Counts the results fulfilling a given predicate.
   * @param pred the predicate to test
   * @return the number of student results fulfilling the predicate
   */
  public int count(Predicate<Results> pred) {
    int n = 0;
    for (Results r : resultsList) {
      if (pred.test(r))
        n++;
    }
    return n;
  }

  /**
   * Makes a computation on the student result records.
   * Uses a accumulator and a finalizer function.
   *
   * @param <A> the type of the start and accumulated value
   * @param <R> the type of the result
   * @param start the start value
   * @param accumulator the accumulator function
   * @param finalizer the finalizer function
   * @return the computation result
   */
  public <A, R> R compute(A start, BiFunction<? super A, Results, ? extends A> accumulator,
      Function<? super A, ? extends R> finalizer) {
    A a = resultsList.reduce(start, accumulator);
    return finalizer.apply(a);
  }

}
