package at.jku.ssw.fp.sect04_2.imp;

import java.util.HashMap;
import java.util.Map;

/**
 * Class for representing and maintaining a list of result records
 * for a set of students.
 *
 * @author Herbert Praehofer
 */
public class ResultsList {

  /**
   * Map of student ids to their results.
   */
  private final Map<String, Results> resultsMap;

  /**
   * Constructor initializing the map of results.
   */
  public ResultsList() {
    resultsMap = new HashMap<String, Results>();
  }

  /**
   * Adds a student to the list.
   * @param id the id of the student
   * @param name the name of the student
   */
  public void addStudent(String id, String name) {
    resultsMap.put(id, new Results(id, name));
  }

  /**
   * Sets the score for a student for an assignment.
   *
   * @param id the id of the student
   * @param assignment the assignment number
   * @param score the score
   */
  public void setScore(String id, int assignment, int score) {
    assert (assignment >= 1 && assignment <= 10);
    assert (score >= 0 && score <= 10);
    getResults(id).setScore(assignment, score);
  }

  /**
   * Sets the scores for a student.
   *
   * @param id the id of the student
   * @param scores the scores for the assignments
   */
  public void setScore(String id, int... scores) {
    assert (scores.length == 10);
    getResults(id).setScores(scores);
  }

  /**
   * Gets the result object for a student.
   *
   * @param id the id of the student
   * @return the result object
   */
  Results getResults(String id) {
    assert (resultsMap.containsKey(id));
    return resultsMap.get(id);
  }

  /**
   * Gets the score of a student for a particular assignment.
   * @param id the student id
   * @param assignment the assignment number
   * @return the score
   */
  public int getScore(String id, int assignment) {
    assert (assignment >= 1 && assignment <= 10);
    return getResults(id).getScore(assignment);
  }

  /**
   * Counts the assignments of a student which are solved.
   * An assignment is solved if 40% of the points have been achieved.
   *
   * @param id the student id
   * @return the number of solved assignments
   */
  public int countSolved(String id) {
    assert (resultsMap.containsKey(id));
    return getResults(id).countSolved();
  }

  /**
   * Computes the average of the scores of a student.
   *
   * @param id the student id
   * @return the average of the scores of the student
   */
  public double avrgScores(String id) {
    return getResults(id).avrgScores();
  }

  /**
   * Computes the grade from the scores for the given student.
   *
   * @param id the student id
   * @return the grade
   */
  public Grade getGrade(String id) {
    return getResults(id).getGrade();
  }

  /**
   * Counts the number of students which have solved the given assignment.
   *
   * @param assignment the number of the assignment
   * @return the number of students solved the assignment
   */
  public int countSolved(int assignment) {
    int n = 0;
    for (Results r : resultsMap.values()) {
      if (r.getScore(assignment) >= 4) {
        n++;
      }
    }
    return n;
  }

  /**
   * Computes the average of points achieved in a particular assignment.
   *
   * @param assignment the number of the assignment
   * @return the average for this assignment
   */
  public double avrgScores(int assignment) {
    int sum = 0;
    for (String stdtId : resultsMap.keySet()) {
      sum += getScore(stdtId, assignment);
    }
    return (double) sum / resultsMap.size();
  }

}
