package at.jku.ssw.fp.sect04_3;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Optional;
import java.util.stream.Collectors;

/**
 * Class representing a node in a graph. 
 * A node has a value. 
 * 
 * @author Herbert Praehofer 
 *
 * @param <V> the type of the value of the nodes
 */
public class Node<V> {

  /**
   * Connects two nodes. 
   * 
   * @param <V> the type of node values
   * @param from the source node
   * @param to the target node
   * @param weight the weight
   * @return the edge
   */
  static <V> Edge<V> connect(Node<V> from, Node<V> to, int weight) {
    Edge<V> edge = new Edge<V>(from, to, weight);
    from.succs.add(edge);
    to.preds.add(edge);
    return edge;
  }

  /** The value of the node */
  private final V value;
  
  /** The list of outgoing edges */
  private final List<Edge<V>> succs;
  
  /** The list of incoming edges */
  private final List<Edge<V>> preds;

  /**
   * Constructor initializing the value of the node. 
   * @param value the value of the node 
   */
  Node(V value) {
    super();
    this.value = value;
    this.succs = new ArrayList<>();
    this.preds = new ArrayList<>();
  }

  /** 
   * Returns the value of the node. 
   * 
   * @return the value of the node
   */
  public V getValue() {
    return value;
  }

  /**
   * Returns the edge building a connection to the given node in an optional;
   * with empty if no connection exists. 
   * 
   * @param succNode the connected node 
   * @return an edge in an optional building the connection
   */
  public Optional<Edge<V>> getEdge(Node<V> succNode) {
    for (Edge<V> e : getSuccEdges()) {
      if (e.getTo().equals(succNode)) {
        return Optional.of(e);
      }
    }
    return Optional.empty();
  }

  /**
   * Returns the weight of the edge building a connection 
   * to the given node in an optional; with empty if no connection exists. 
   * 
   * @param succNode the connected node 
   * @return the weight of an edge in an optional
   */
  public int getWeight(Node<V> succNode) {
    Optional<Edge<V>> optEdge = getEdge(succNode);
    return optEdge.isPresent() ? optEdge.get().getWeight() : Integer.MAX_VALUE;
  }

  /**
   * Returns a list of outgoing edges. 
   * @return the outgoing edges
   */
  public List<Edge<V>> getSuccEdges() {
    return Collections.unmodifiableList(succs);
  }

  /**
   * Returns a list of incoming edges. 
   * @return the incoming edges
   */
  public List<Edge<V>> getPredEdges() {
    return Collections.unmodifiableList(preds);
  }

  /**
   * Returns a list of connected nodes by outgoing edges. 
   * @return the nodes connected by an outgoing edges
   */
  public List<Node<V>> getSuccNodes() {
    return succs.stream().map(e -> e.getTo()).collect(Collectors.toList());
  }

  /**
   * Returns a list of connected nodes by incoming edges. 
   * @return the nodes connected by an incoming edges
   */
  public List<Node<V>> getPredNodes() {
    return succs.stream().map(e -> e.getFrom()).collect(Collectors.toList());
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + ((value == null) ? 0 : value.hashCode());
    return result;
  }

  @SuppressWarnings("unchecked")
  @Override
  public boolean equals(Object obj) {
    if (this == obj)
      return true;
    if (obj == null)
      return false;
    if (getClass() != obj.getClass())
      return false;
    Node<V> other = (Node<V>) obj;
    if (value == null) {
      if (other.value != null)
        return false;
    } else if (!value.equals(other.value))
      return false;
    return true;
  }

  @Override
  public String toString() {
    return String.valueOf(value);
  }

}
