package at.jku.ssw.fp.sect04_3;

import java.util.ArrayDeque;
import java.util.Comparator;
import java.util.Deque;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.PriorityQueue;
import java.util.Queue;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Predicate;

/**
 * Class with search methods. 
 * 
 * @author Herbert Praehofer 
 */
public class SearchAlgos {

  private static final boolean DEBUG = false;
  public static int nodesVisited = 0;

  // depth first search

  /**
   * Implements a depth-first search algorithm.
   * 
   * @param <N> the type of nodes
   * @param succFn the successor function 
   * @param goalFn the goal function 
   * @param start the start node
   * @return the found node in an optional, if exists, empty otherwise. 
   */
  public static <N> Optional<N> depthFirstSearch(Function<N, List<N>> succFn, Predicate<N> goalFn, N start) {
    Set<N> visited = new HashSet<N>();
    Deque<N> stack = new ArrayDeque<N>();
    stack.push(start);
    while (!stack.isEmpty()) {
      N node = stack.pop();
      if (goalFn.test(node)) {
        return Optional.of(node);
      } else {
        visited.add(node);
        List<N> succs = succFn.apply(node);
        succs.removeAll(visited);
        for (N n : succs) {
          if (stack.contains(n))
            stack.remove(n);
          stack.push(n);
        }
      }
    }
    return Optional.empty();
  }

  // breath first search

  /**
   * Implements a breadth-first search algorithm.
   * @param <N> the type of nodes
   * @param succFn the successor function 
   * @param goalFn the goal function 
   * @param start the start node
   * @return the found node in an optional, if exists, empty otherwise. 
   */
  public static <N> Optional<N> breadthFirstSearch(Function<N, List<N>> succFn, Predicate<N> goalFn, N start) {
    Set<N> visited = new HashSet<>();
    Queue<N> queue = new ArrayDeque<N>();
    queue.add(start);
    while (!queue.isEmpty()) {
      N node = queue.remove();
      if (DEBUG)
        System.out.println("Node: " + String.valueOf(node));
      if (goalFn.test(node)) {
        return Optional.of(node);
      } else {
        visited.add(node);
        List<N> succs = succFn.apply(node);
        succs.removeAll(visited);
        if (DEBUG)
          System.out.println("  Succ: " + succs);
        queue.addAll(succs);
      }
    }
    return Optional.empty();
  }

  // best first search

  /**
   * Implements a best-first search algorithm.
   * @param <N> the type of nodes
   * @param succFn the successor function 
   * @param weightingFn the function for ordering the nodes
   * @param goalFn the goal function 
   * @param start the start node
   * @return the found node in an optional, if exists, empty otherwise. 
   */
  public static <N> Optional<N> bestFirstSearch(Function<N, List<N>> succFn, Function<N, Integer> weightingFn,
      Predicate<N> goalFn, N start) {
    final PriorityQueue<N> priorityQueue = new PriorityQueue<N>(
        Comparator.comparingInt(node -> weightingFn.apply(node)));
    final Set<N> visited = new HashSet<>();
    priorityQueue.add(start);
    while (!priorityQueue.isEmpty()) {
      N node = priorityQueue.remove();
      if (DEBUG)
        System.out.println("Node: " + String.valueOf(node));
      if (goalFn.test(node)) {
        return Optional.of(node);
      } else {
        visited.add(node);
        List<N> succs = succFn.apply(node);
        if (DEBUG)
          System.out.println("  Succ: " + succs);
        for (N succ : succs) {
          if (!visited.contains(succ)) {
            priorityQueue.add(succ);
          }
        }
      }
    }
    return Optional.empty();
  }

}
