package at.jku.ssw.fp.sect04_3;

import java.util.ArrayDeque;
import java.util.Collection;
import java.util.Comparator;
import java.util.Deque;
import java.util.HashSet;
import java.util.List;
import java.util.Optional;
import java.util.PriorityQueue;
import java.util.Queue;
import java.util.Set;
import java.util.function.BiConsumer;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.function.ToIntFunction;

/**
 * Class implementing the generic version of the search algorithm. 
 * 
 * @author Herbert Praehofer 
 *
 */
public class SearchAlgosGeneric {

  public static int nodesVisited = 0;

  /**
   * The interface for the controller of the search algorithm. 
   * 
   * @param <N> the type of the nodes
   * @param <C> the type of the collection used in the search algorithm
   */
  public interface Controller<N, C extends Collection<N>> {
    /**
     * Returns the supplier function to create the collection used in the search algorithm.
     *  
     * @return the supplier function for the collection
     */
    public Supplier<C> createFn();

    /**
     * Returns the function to retrieve the next node to handle from the collection. 
     * @return the function to retrieve the next node
     */
    public Function<C, N> nextNodeFn();

    /**
     * Returns the function to update the collection given a list of next nodes. 
     * @return the function to update the collection
     */
    public BiConsumer<List<N>, C> updateFn();
  }

  /**
   * Implements a generic search algorithm. 
   * 
   * @param <N> the type of the nodes
   * @param <C> the type of the collection used in the search algorithm
   * @param controller the controller object for controlling the search
   * @param successorFn the function for the successors nodes of a node
   * @param goalFn the goal function 
   * @param start the start node
   * @return the found node in an optional, if exists, empty otherwise. 
   */
  public static <N, C extends Collection<N>> Optional<N> search(Controller<N, C> controller,
      Function<N, List<N>> successorFn, Predicate<N> goalFn, N start) {
    Set<N> visited = new HashSet<>();
    C coll = controller.createFn().get();
    coll.add(start);
    while (!coll.isEmpty()) {
      N node = controller.nextNodeFn().apply(coll);
      if (goalFn.test(node)) {
        return Optional.of(node);
      } else {
        visited.add(node);
        List<N> succs = successorFn.apply(node);
        succs.removeAll(visited);
        succs.removeAll(coll);
        controller.updateFn().accept(succs, coll);
      }
    }
    return Optional.empty();
  }

  // depth first search

  /**
   * Implements a depth-first search algorithm by the generic algorithm 
   * using a special controller.
   * 
   * @param <N> the type of nodes
   * @param succFn the successor function 
   * @param goalFn the goal function 
   * @param start the start node
   * @return the found node in an optional, if exists, empty otherwise. 
   */
  public static <N> Optional<N> depthFirstSearch(Function<N, List<N>> succFn, Predicate<N> goalFn, N start) {
    return search(new Controller<N, Deque<N>>() {
      @Override
      public Supplier<Deque<N>> createFn() {
        return () -> new ArrayDeque<N>();
      }

      @Override
      public Function<Deque<N>, N> nextNodeFn() {
        return stack -> stack.pop();
      }

      @Override
      public BiConsumer<List<N>, Deque<N>> updateFn() {
        return (nextNodes, stack) -> {
          for (N n : nextNodes)
            stack.push(n);
        };
      }
    }, succFn, goalFn, start);
  }

  // breath first search

  /**
   * Implements a breadth-first search algorithm by the generic algorithm 
   * using a special controller.
   * 
   * @param <N> the type of nodes
   * @param succFn the successor function 
   * @param goalFn the goal function 
   * @param start the start node
   * @return the found node in an optional, if exists, empty otherwise. 
   */
  public static <N> Optional<N> breadthFirstSearch(Function<N, List<N>> succFn, Predicate<N> goalFn, N start) {
    return search(new Controller<N, Queue<N>>() {
      @Override
      public Supplier<Queue<N>> createFn() {
        return () -> new ArrayDeque<N>();
      }

      @Override
      public Function<Queue<N>, N> nextNodeFn() {
        return queue -> queue.remove();
      }

      @Override
      public BiConsumer<List<N>, Queue<N>> updateFn() {
        return (nextNodes, queue) -> queue.addAll(nextNodes);
      }
    }, succFn, goalFn, start);
  }

  // priority first search

  /**
   * Implements a best-first search algorithm by the generic algorithm 
   * using a special controller.
   * 
   * @param <N> the type of nodes
   * @param succFn the successor function 
   * @param weightingFn the function for ordering the nodes
   * @param goalFn the goal function 
   * @param start the start node
   * @return the found node in an optional, if exists, empty otherwise. 
   */
  public static <N> Optional<N> bestFirstSearch(Function<N, List<N>> succFn, ToIntFunction<N> weightingFn, Predicate<N> goalFn,
      N start) {
    return search(new Controller<N, PriorityQueue<N>>() {
      @Override
      public Supplier<PriorityQueue<N>> createFn() {
        return () -> {
          PriorityQueue<N> queue = new PriorityQueue<N>(Comparator.comparingInt(n -> weightingFn.applyAsInt(n)));
          return queue;
        };
      }

      @Override
      public Function<PriorityQueue<N>, N> nextNodeFn() {
        return queue -> queue.remove();
      }

      @Override
      public BiConsumer<List<N>, PriorityQueue<N>> updateFn() {
        return (nextNodes, queue) -> {
          for (N node : nextNodes) {
            if (!queue.contains(node))
              queue.add(node);
          }
        };
      }
    }, succFn, goalFn, start);
  }


}
