package at.jku.ssw.fp.sect04_3.puzzle;

import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.List;
import java.util.Random;
import java.util.stream.Collectors;

/**
 * State in a puzzle game.
 *
 * @author Herbert Praehofer
 *
 */
public final class Puzzle {

  /** Size of the puzzle board */
  public static final int N = 3;

  /** A random number generator */
  static final Random RAND = new Random();

  /** Array representing solution */
  static final int[] NUMBERS = new int[N * N];
  {
    for (int i = 0; i < NUMBERS.length; i++) {
      NUMBERS[i] = (i + 1) % (N * N);
    }
  }

  /** Solution state */
  public static final Puzzle SOLUTION = new Puzzle(NUMBERS);

  /**
   * Creates a random puzzle state.
   *
   * @return a randomly set puzzle state
   */
  public static Puzzle shuffled() {
    Puzzle p = SOLUTION;
    while (p.assess() > 0) {
      List<Move> moves = p.getNextMoves();
      int i = RAND.nextInt(moves.size());
      p = p.doMove(moves.get(i));
    }
    for (int x = 0; x < 100; x++) {
      List<Move> moves = p.getNextMoves();
      int i = RAND.nextInt(moves.size());
      p = p.doMove(moves.get(i));
    }
    return new Puzzle(p.getNumbers());
  }

  /**
   * Creates a puzzle state from sequence of numbers.
   *
   * @param numbers the sequence of numbers
   * @return the puzzle from the sequence of numbers
   */
  public static Puzzle from(int... numbers) {
    return new Puzzle(numbers);
  }

  /**
   * Gets the number at the current position.
   *
   * @param pos current position
   * @return the number for the current position
   */
  public int get(Pos pos) {
    return get(pos.row, pos.col);
  }

  /**
   * Tests if this puzzle state represents the solution state.
   *
   * @return {@code true} if this represents the solution state
   */
  public boolean isSolution() {
    return this.equals(SOLUTION);
  }

  /**
   * Returns a list of possible successor states
   *
   * @return the list of successor states
   */
  public List<Puzzle> getNextStates() {
    return getNextMoves().stream().map(move -> this.doMove(move)).collect(Collectors.toList());
  }

  /**
   * Computes a difference of this state to the solution state.
   *
   * @return the difference to the solution state
   */
  public int assess() {
    int n = 0;
    for (int i = 0; i < field.length; i++) {
      for (int j = 0; j < field[i].length; j++) {
        if (field[i][j] == SOLUTION.get(i, j)) {
          n++;
        }
      }
    }
    return n;
  }

  /**
   * Gets the position which is empty.
   *
   * @return the empty position
   */
  public Pos getEmptyPos() {
    for (int i = 0; i < field.length; i++) {
      for (int j = 0; j < field[i].length; j++) {
        if (get(i, j) == 0) {
          return Pos.of(i, j);
        }
      }
    }
    return null;
  }

  /**
   * Returns the sequence of numbers in a list.
   *
   * @return the sequence of numbers
   */
  public List<Integer> getNumbers() {
    List<Integer> ns = new ArrayList<>();
    for (int i = 0; i < field.length; i++) {
      for (int j = 0; j < field[i].length; j++) {
        ns.add(field[i][j]);
      }
    }
    return ns;
  }

  /**
   * Gets the list of moves  taken to get to this state.
   *
   * @return a list of moves taken to get to this state
   */
  public List<Move> getMoveHistory() {
    return Collections.unmodifiableList(moves);
  }

  /**
   * Gets the list of states to this state.
   *
   * @return a list of states to this state
   */
  public List<Puzzle> getStateHistory() {
    return Collections.unmodifiableList(states);
  }

  /**
   * Gets the moves possible from this state.
   *
   * @return the list of possible moves
   */
  public List<Move> getNextMoves() {
    Pos empty = getEmptyPos();
    List<Move> moves = new ArrayList<Move>();
    for (int i = empty.row - 1; i <= empty.row + 1; i++) {
      Pos from = Pos.of(i, empty.col);
      if (from.isValid() && i != empty.row) {
        moves.add(Move.of(from, empty));
      }
    }
    for (int j = empty.col - 1; j <= empty.col + 1; j++) {
      Pos from = Pos.of(empty.row, j);
      if (from.isValid() && j != empty.col) {
        moves.add(Move.of(from, empty));
      }
    }
    return Collections.unmodifiableList(moves);
  }

  /**
   * Executes a next move.
   *
   * @param move the move
   * @return the next puzzle state
   */
  public Puzzle doMove(Move move) {
    Pos empty = getEmptyPos();
    int temp = get(move.from);
    Puzzle moved = new Puzzle();
    moved.moves = new ArrayList<>();
    moved.moves.addAll(this.moves);
    moved.moves.add(move);
    moved.states = new ArrayList<>();
    moved.states.addAll(this.states);
    moved.states.add(this);
    for (int i = 0; i < field.length; i++) {
      for (int j = 0; j < field[i].length; j++) {
        Pos pos = Pos.of(i, j);
        if (pos.equals(empty)) {
          moved.field[i][j] = temp;
        } else if (pos.equals(move.from)) {
          moved.field[i][j] = 0;
        } else {
          moved.field[i][j] = this.field[i][j];
        }
      }
    }
    return moved;
  }

  @Override
  public int hashCode() {
    final int prime = 31;
    int result = 1;
    result = prime * result + Arrays.deepHashCode(field);
    return result;
  }

  @Override
  public boolean equals(Object obj) {
    if (this == obj)
      return true;
    if (obj == null)
      return false;
    if (getClass() != obj.getClass())
      return false;
    Puzzle other = (Puzzle) obj;
    if (!equalFieds(field, other.field))
      return false;
    return true;
  }

  @Override
  public String toString() {
    StringBuilder sb = new StringBuilder();
    for (int i = 0; i < field.length; i++) {
      for (int j = 0; j < field[i].length; j++) {
        int x = field[i][j];
        if (x != 0) {
          sb.append(String.format(" %2d ", x));
        } else {
          sb.append("  - ");
        }
      }
      sb.append("\n");
    }
    return sb.toString();
  }

  public void print(PrintStream out) {
    out.println(toString());
  }

  // private -----------------

  private final int[][] field = new int[N][N];
  private List<Move> moves;
  private List<Puzzle> states;

  private Puzzle() {
    moves = new ArrayList<>();
    states = new ArrayList<>();
  }

  private Puzzle(int... numbers) {
    for (int i = 0; i < numbers.length; i++) {
      field[i / N][i % N] = numbers[i];
    }
    moves = new ArrayList<>();
    states = new ArrayList<>();
  }

  private Puzzle(List<Integer> numbers) {
    for (int i = 0; i < numbers.size(); i++) {
      field[i / N][i % N] = numbers.get(i);
    }
    moves = new ArrayList<>();
    states = new ArrayList<>();
  }

  private boolean equalFieds(int[][] f1, int[][] f2) {
    for (int i = 0; i < f1.length; i++) {
      for (int j = 0; j < f1[i].length; j++) {
        if (f1[i][j] != f2[i][j])
          return false;
      }
    }
    return true;
  }

  private int get(int i, int j) {
    return field[i][j];
  }

}
