package at.jku.ssw.fp.sect04_4.visitor_fp;

import java.util.function.Consumer;
import java.util.function.Function;

/**
 * Abstract base class of Boolean expressions.
 * Contains static factory methods for creating the various
 * Boolean expressions.
 *
 * @author Herbert Praehofer
 *
 */
abstract public class Expr {

  /**
   * Creates an literal expression with a value.
   *
   * @param value the Boolean value of the literal
   * @return the literal expression
   */
  public static Lit mkLit(boolean value) {
    return new Lit(value);
  }

  /**
   * Creates an variable expression with a name.
   *
   * @param name the variable name
   * @return the variable expression
   */
  public static Var mkVar(String name) {
    return new Var(name);
  }

  /**
   * Returns an and-expression with a left and right subexpression.
   *
   * @param left the left subexpression
   * @param right the right subexpression
   * @return the and-expression
   */
  public static And mkAnd(Expr left, Expr right) {
    return new And(left, right);
  }

  /**
   * Creates an or-expression with a left and right subexpression.
   *
   * @param left the left subexpression
   * @param right the right subexpression
   * @return the or-expression
   */
  public static Or mkOr(Expr left, Expr right) {
    return new Or(left, right);
  }

  /**
   * Creates an not-expression with a subexpression.
   *
   * @param expr the subexpression
   * @return the not-expression
   */
  public static Not mkNot(Expr expr) {
    return new Not(expr);
  }

  /**
   * Applies one of the given functions dependent on the concrete
   * expression type.
   *
   * @param <R> the type of the return value
   * @param litFn the function for {@link Lit} objects
   * @param varFn the function for {@link Var} objects
   * @param andFn the function for {@link And} objects
   * @param orFn the function for  {@link Or} objects
   * @param notFn the function for {@link Not} objects
   * @return the result of the function application
   */
  public abstract <R> R apply(
      Function<Lit, R> litFn,
      Function<Var, R> varFn,
      Function<And, R> andFn,
      Function<Or, R> orFn,
      Function<Not, R> notFn
    );

  /**
   * Executes one of the given actions dependent on the concrete
   * expression type.
   *
   * @param litFn the action for {@link Lit} objects
   * @param varFn the action for {@link Var} objects
   * @param andFn the action for {@link And} objects
   * @param orFn the action for  {@link Or} objects
   * @param notFn the action for {@link Not} objects
   */
  public abstract void accept(
      Consumer<Lit> litFn,
      Consumer<Var> varFn,
      Consumer<And> andFn,
      Consumer<Or> orFn,
      Consumer<Not> notFn);

  /**
   * Class representing literal expressions.
   */
  public static class Lit extends Expr {

    /** The value of this literal expression */
    public final boolean value;

    /**
     * Constructor for literals.
     * @param value the value of the literal
     */
    public Lit(boolean value) {
      this.value = value;
    }

    /**
     * Returns a string representation.
     *
     * @return the string representation
     */
    @Override
    public String toString() {
      return String.valueOf(value);
    }

    /**
     * Applies the function for the {@link Lit} type.
     *
     * @param <R> the type of the return value
     * @param litFn the function for {@link Lit} objects
     * @param varFn the function for {@link Var} objects
     * @param andFn the function for {@link And} objects
     * @param orFn the function for  {@link Or} objects
     * @param notFn the function for {@link Not} objects
     * @return the result of the function application
     */
    @Override
    public <R> R apply(
        Function<Lit, R> litFn,
        Function<Var, R> varFn,
        Function<And, R> andFn,
        Function<Or, R> orFn,
        Function<Not, R> notFn
      ) {
      return litFn.apply(this);
    }


    /**
     * Executes the action for the {@link Lit} type.
     *
     * @param litFn the action for {@link Lit} objects
     * @param varFn the action for {@link Var} objects
     * @param andFn the action for {@link And} objects
     * @param orFn the action for  {@link Or} objects
     * @param notFn the action for {@link Not} objects
     */
    @Override
    public void accept(
        Consumer<Lit> litFn,
        Consumer<Var> varFn,
        Consumer<And> andFn,
        Consumer<Or> orFn,
        Consumer<Not> notFn
      ) {
      litFn.accept(this);
    }

  }

  /**
   * The class representing variable expressions.
   */
  public static class Var extends Expr {

    /** The variable name */
    public final String name;

    /**
     * Constructor for variable expressions.
     *
     * @param name the variable name
     */
    public Var(String name) {
      super();
      this.name = name;
    }

    /**
     * Returns a string representation.
     *
     * @return the string representation
     */
    @Override
    public String toString() {
      return name;
    }

    /**
     * Applies the function for the {@link Var} type.
     *
     * @param <R> the type of the return value
     * @param litFn the function for {@link Lit} objects
     * @param varFn the function for {@link Var} objects
     * @param andFn the function for {@link And} objects
     * @param orFn the function for  {@link Or} objects
     * @param notFn the function for {@link Not} objects
     * @return the result of the function application
     */
    @Override
    public <R> R apply(
        Function<Lit, R> litFn,
        Function<Var, R> varFn,
        Function<And, R> andFn,
        Function<Or, R> orFn,
        Function<Not, R> notFn
      ) {
      return varFn.apply(this);
    }

    /**
     * Executes the action for the {@link Var} type.
     *
     * @param litFn the action for {@link Lit} objects
     * @param varFn the action for {@link Var} objects
     * @param andFn the action for {@link And} objects
     * @param orFn the action for  {@link Or} objects
     * @param notFn the action for {@link Not} objects
     */
    @Override
    public void accept(
        Consumer<Lit> litFn,
        Consumer<Var> varFn,
        Consumer<And> andFn,
        Consumer<Or> orFn,
        Consumer<Not> notFn
      ) {
      varFn.accept(this);
    }

  }

  /**
   * Abstract base class for binary expressions.
   */
  static abstract class BinOp extends Expr {

    /** The left subexpression of the binary expression */
    public final Expr left;

    /** The right subexpression of the binary expression */
    public final Expr right;

    /**
     * Constructor initializing left and right subexpressions.
     *
     * @param left the left subexpression
     * @param right the right subexpression
     */
    public BinOp(Expr left, Expr right) {
      this.left = left;
      this.right = right;
    }

    /**
     * Returns a string representation.
     *
     * @return the string representation
     */
    @Override
    public String toString() {
      return "(" + left.toString() + getOp() + right.toString() + ")";
    }

    /**
     * Returns the string for the operator of the binary expression.
     *
     * @return the string for the operator
     */
    protected abstract String getOp();

  }

  /**
   * Class representing and-expressions.
   */
  public static class And extends BinOp {

    /**
     * Constructor initializing left and right subexpressions.
     *
     * @param left the left subexpression
     * @param right the right subexpression
     */
    public And(Expr left, Expr right) {
      super(left, right);
    }

    /**
     * Returns {@code " && "} as string for the operator.
     *
     * @return {@code " && "}
     */
    @Override
    protected String getOp() {
      return " && ";
    }

    /**
     * Applies the function for the {@link And} type.
     *
     * @param <R> the type of the return value
     * @param litFn the function for {@link Lit} objects
     * @param varFn the function for {@link Var} objects
     * @param andFn the function for {@link And} objects
     * @param orFn the function for  {@link Or} objects
     * @param notFn the function for {@link Not} objects
     * @return the result of the function application
     */
    @Override
    public <R> R apply(
        Function<Lit, R> litFn,
        Function<Var, R> varFn,
        Function<And, R> andFn,
        Function<Or, R> orFn,
        Function<Not, R> notFn
      ) {
      return andFn.apply(this);
    }

    /**
     * Executes the action for the {@link And} type.
     *
     * @param litFn the action for {@link Lit} objects
     * @param varFn the action for {@link Var} objects
     * @param andFn the action for {@link And} objects
     * @param orFn the action for  {@link Or} objects
     * @param notFn the action for {@link Not} objects
     */
    @Override
    public void accept(
        Consumer<Lit> litFn,
        Consumer<Var> varFn,
        Consumer<And> andFn,
        Consumer<Or> orFn,
        Consumer<Not> notFn
      ) {
      andFn.accept(this);
    }

  }

  /**
   * Class representing or-expressions.
   */
  public static class Or extends BinOp {

    /**
     * Constructor initializing left and right subexpressions.
     *
     * @param left the left subexpression
     * @param right the right subexpression
     */
    public Or(Expr left, Expr right) {
      super(left, right);
    }

    /**
     * Returns {@code " || "} as string for the operator.
     *
     * @return {@code " || "}
     */
    @Override
    protected String getOp() {
      return " || ";
    }

    /**
     * Applies the function for the {@link Or} type.
     *
     * @param <R> the type of the return value
     * @param litFn the function for {@link Lit} objects
     * @param varFn the function for {@link Var} objects
     * @param andFn the function for {@link And} objects
     * @param orFn the function for  {@link Or} objects
     * @param notFn the function for {@link Not} objects
     * @return the result of the function application
     */
    @Override
    public <R> R apply(
        Function<Lit, R> litFn,
        Function<Var, R> varFn,
        Function<And, R> andFn,
        Function<Or, R> orFn,
        Function<Not, R> notFn
      ) {
      return orFn.apply(this);
    }

    /**
     * Executes the action for the {@link Or} type.
     *
     * @param litFn the action for {@link Lit} objects
     * @param varFn the action for {@link Var} objects
     * @param andFn the action for {@link And} objects
     * @param orFn the action for  {@link Or} objects
     * @param notFn the action for {@link Not} objects
     */
    @Override
    public void accept(
        Consumer<Lit> litFn,
        Consumer<Var> varFn,
        Consumer<And> andFn,
        Consumer<Or> orFn,
        Consumer<Not> notFn
      ) {
      orFn.accept(this);
    }

  }

  /**
   * Class representing not-expressions.
   */
  public static class Not extends Expr {

    /** The subexpression of this not-expression. */
    public final Expr sub;

    /**
     * Constructor initializing the subexpression.
     *
     * @param sub the subexpression
     */
    public Not(Expr sub) {
      this.sub = sub;
    }

    /**
     * Returns a string representation.
     *
     * @return the string representation
     */
    @Override
    public String toString() {
      return "(! " + sub.toString() + ")";
    }

    /**
     * Applies the function for the {@link Not} type.
     *
     * @param <R> the type of the return value
     * @param litFn the function for {@link Lit} objects
     * @param varFn the function for {@link Var} objects
     * @param andFn the function for {@link And} objects
     * @param orFn the function for  {@link Or} objects
     * @param notFn the function for {@link Not} objects
     * @return the result of the function application
     */
    @Override
    public <R> R apply(
        Function<Lit, R> litFn,
        Function<Var, R> varFn,
        Function<And, R> andFn,
        Function<Or, R> orFn,
        Function<Not, R> notFn
      ) {
      return notFn.apply(this);
    }

    /**
     * Executes the action for the {@link Not} type.
     *
     * @param litFn the action for {@link Lit} objects
     * @param varFn the action for {@link Var} objects
     * @param andFn the action for {@link And} objects
     * @param orFn the action for  {@link Or} objects
     * @param notFn the action for {@link Not} objects
     */
    @Override
    public void accept(
        Consumer<Lit> litFn,
        Consumer<Var> varFn,
        Consumer<And> andFn,
        Consumer<Or> orFn,
        Consumer<Not> notFn
      ) {
      notFn.accept(this);
    }

  }

}
