package at.jku.ssw.fp.sect04_4.visitor_nofp;

/**
 * Abstract base class of Boolean expressions.
 * Contains static factory methods for creating the various
 * Boolean expressions.
 *
 * @author Herbert Praehofer
 *
 */
abstract public class Expr {

  /**
   * Creates an literal expression with a value.
   *
   * @param value the Boolean value of the literal
   * @return the literal expression
   */
  public static Lit mkLit(boolean value) {
    return new Lit(value);
  }

  /**
   * Creates an variable expression with a name.
   *
   * @param name the variable name
   * @return the variable expression
   */
  public static Var mkVar(String name) {
    return new Var(name);
  }

  /**
   * Returns an and-expression with a left and right subexpression.
   *
   * @param left the left subexpression
   * @param right the right subexpression
   * @return the and-expression
   */
  public static And mkAnd(Expr left, Expr right) {
    return new And(left, right);
  }

  /**
   * Creates an or-expression with a left and right subexpression.
   *
   * @param left the left subexpression
   * @param right the right subexpression
   * @return the or-expression
   */
  public static Or mkOr(Expr left, Expr right) {
    return new Or(left, right);
  }

  /**
   * Creates an not-expression with a subexpression.
   *
   * @param expr the subexpression
   * @return the not-expression
   */
  public static Not mkNot(Expr expr) {
    return new Not(expr);
  }

  /**
   * Accepts a visitor.
   *
   * @param <R> the type of the return value of the visitor.
   *
   * @param v the visitor
   * @return the return value
   */
  public abstract <R> R accept(Visitor<R> v);


  /**
   * Class representing literal expressions.
   */
  public static class Lit extends Expr {

    /** The value of this literal expression */
    public final boolean value;

    /**
     * Constructor for literals.
     * @param value the value of the literal
     */
    public Lit(boolean value) {
      this.value = value;
    }

    /**
     * Returns a string representation.
     *
     * @return the string representation
     */
    @Override
    public String toString() {
      return String.valueOf(value);
    }

    /**
     * Accepts a visitor by calling the {@link Visitor#visitLit(Lit)} method.
     *
     * @param <R> the type of the return value of the visitor.
     *
     * @param v the visitor
     * @return the return value
     */
    @Override
    public <R> R accept(Visitor<R> v) {
      return v.visitLit(this);
    }

  }

  /**
   * The class representing variable expressions.
   */
  public static class Var extends Expr {

    /** The variable name */
    public final String name;

    /**
     * Constructor for variable expressions.
     *
     * @param name the variable name
     */
    public Var(String name) {
      super();
      this.name = name;
    }

    /**
     * Returns a string representation.
     *
     * @return the string representation
     */
    @Override
    public String toString() {
      return name;
    }

    /**
     * Accepts a visitor by calling the {@link Visitor#visitVar(Var)} method.
     *
     * @param <R> the type of the return value of the visitor.
     *
     * @param v the visitor
     * @return the return value
     */
    @Override
    public <R> R accept(Visitor<R> v) {
      return v.visitVar(this);
    }

  }

  /**
   * Abstract base class for binary expressions.
   */
  static abstract class BinOp extends Expr {

    /** The left subexpression of the binary expression */
    public final Expr left;

    /** The right subexpression of the binary expression */
    public final Expr right;

    /**
     * Constructor initializing left and right subexpressions.
     *
     * @param left the left subexpression
     * @param right the right subexpression
     */
    public BinOp(Expr left, Expr right) {
      this.left = left;
      this.right = right;
    }

    /**
     * Returns a string representation.
     *
     * @return the string representation
     */
    @Override
    public String toString() {
      return "(" + left.toString() + getOp() + right.toString() + ")";
    }

    /**
     * Returns the string for the operator of the binary expression.
     *
     * @return the string for the operator
     */
    protected abstract String getOp();

  }

  /**
   * Class representing and-expressions.
   */
  public static class And extends BinOp {

    /**
     * Constructor initializing left and right subexpressions.
     *
     * @param left the left subexpression
     * @param right the right subexpression
     */
    public And(Expr left, Expr right) {
      super(left, right);
    }

    /**
     * Returns {@code " && "} as string for the operator.
     *
     * @return {@code " && "}
     */
    @Override
    protected String getOp() {
      return " && ";
    }

    /**
     * Accepts a visitor by calling the {@link Visitor#visitAnd(And)} method.
     *
     * @param <R> the type of the return value of the visitor.
     *
     * @param v the visitor
     * @return the return value
     */
   @Override
    public <R> R accept(Visitor<R> v) {
      return v.visitAnd(this);
    }

  }

  /**
   * Class representing or-expressions.
   */
  public static class Or extends BinOp {

    /**
     * Constructor initializing left and right subexpressions.
     *
     * @param left the left subexpression
     * @param right the right subexpression
     */
    public Or(Expr left, Expr right) {
      super(left, right);
    }

    /**
     * Returns {@code " || "} as string for the operator.
     *
     * @return {@code " || "}
     */
    @Override
    protected String getOp() {
      return " || ";
    }

    /**
     * Accepts a visitor by calling the {@link Visitor#visitOr(Or)} method.
     *
     * @param <R> the type of the return value of the visitor.
     *
     * @param v the visitor
     * @return the return value
     */
    @Override
    public <R> R accept(Visitor<R> v) {
      return v.visitOr(this);
    }

  }

  /**
   * Class representing not-expressions.
   */
  public static class Not extends Expr {

    /** The subexpression of this not-expression. */
    public final Expr sub;

    /**
     * Constructor initializing the subexpression.
     *
     * @param sub the subexpression
     */
    public Not(Expr sub) {
      this.sub = sub;
    }

    /**
     * Returns a string representation.
     *
     * @return the string representation
     */
    @Override
    public String toString() {
      return "(! " + sub.toString() + ")";
    }

    /**
     * Accepts a visitor by calling the {@link Visitor#visitNot(Not)} method.
     *
     * @param <R> the type of the return value of the visitor.
     *
     * @param v the visitor
     * @return the return value
     */
    @Override
    public <R> R accept(Visitor<R> v) {
      return v.visitNot(this);
    }

  }

}
