package at.jku.ssw.fp.sect04_6;

import java.io.IOException;
import java.io.Reader;
import java.util.ArrayDeque;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.Queue;
import java.util.Set;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;
import java.util.function.UnaryOperator;

import at.jku.ssw.fp.util.Ref;

/**
 * Class with lazy methods and lazy iterator and iterable implementations.
 *
 * @author Herbert Prähofer
 *
 */
public class Lazy {

  /**
   * Creates a lazy iterator from the given supplier function.
   *
   * @param <T> the type of the elements
   * @param supplier the supplier function
   * @return the lazy iterator
   */
  public static <T> Iterator<T> iterator(Supplier<Optional<T>> supplier) {
    return new LazyIterator<T>(supplier);
  }

  /**
   * Creates a lazy iterable from the given supplier function. Uses a lazy iterator internally.
   *
   * @param <T> the type of the elements
   * @param supplier the supplier function
   * @return the lazy iterable object

   */
  public static <T> Iterable<T> iterable(Supplier<Optional<T>> supplier) {
    return new Iterable<T>() {
      @Override
      public Iterator<T> iterator() {
        return new LazyIterator<T>(supplier);
      }
    };
  }

  /**
   * Finds an element in the elements from the given iterator
   * which satisfies the given predicate.
   *
   * @param <T> the type of the elements
   * @param iterator the iterator for the elements; possibly lazy
   * @param pred the predicate
   * @return the found element in an optional; or an empty optional if none found
   */
  public static <T> Optional<T> find(Iterator<T> iterator, Predicate<T> pred) {
    while (iterator.hasNext()) {
      T v = iterator.next();
      if (pred.test(v))
        return Optional.of(v);
    }
    return Optional.empty();
  }

  /**
   * Finds an element in the elements in the given iterable
   * which satisfies the given predicate.
   *
   * @param <T> the type of the elements
   * @param iterable the iterable with the elements; possibly lazy
   * @param pred the predicate
   * @return the found element in an optional; or an empty optional if none found
   */
  public static <T> Optional<T> find(Iterable<T> iterable, Predicate<T> pred) {
    for (T t : iterable) {
      if (pred.test(t))
        return Optional.of(t);
    }
    return Optional.empty();
  }

  /**
   * Creates an iterable the element of which are generated by the supplier function.
   *
   * @param <T> the type of the elements
   * @param supplier the supplier function
   * @return the iterable with the generated elements
   */
  public static <T> Iterable<T> generatable(Supplier<T> supplier) {
    return iterable(() -> Optional.of(supplier.get()));
  }

  /**
   * Creates an iterator for the elements which are generated by the supplier function.
   *
   * @param <T> the type of the elements
   * @param supplier the supplier function
   * @return the iterator for the generated elements
   */
  public static <T> Iterator<T> generate(Supplier<T> supplier) {
    return iterator(() -> Optional.of(supplier.get()));
  }

  /**
   * Creates an iterator for the elements which are generated
   * by the unary operator starting from an initial value.
   *
   * @param <T> the type of the elements
   * @param initial the initial value
   * @param nextFn the unary function for creating a next value from the current value
   * @return the iterator
   */
  public static <T> Iterator<T> iterator(T initial, UnaryOperator<T> nextFn) {
    Ref<T> next = Ref.of(initial);
    return iterator(() -> {
      T value = next.value;
      next.value = nextFn.apply(value);
      return Optional.of(value);
    });
  }

  /**
   * Creates an iterable with elements which are generated
   * by the unary operator starting from an initial value.
   *
   * @param <T> the type of the elements
   * @param initial the initial value
   * @param nextFn the unary function for creating a next value from the current value
   * @return the iterable
   */
  public static <T> Iterable<T> iterate(T initial, UnaryOperator<T> nextFn) {
    Ref<T> next = Ref.of(initial);
    return Lazy.iterable(() -> {
      T value = next.value;
      next.value = nextFn.apply(value);
      return Optional.of(value);
    });
  }

//  /**
//   *
//   * @param <T>
//   * @param iterable
//   * @return
//   */
//  public static <T> Iterable<T> fromIterable(Iterable<T> iterable) {
//    Iterator<T> it = iterable.iterator();
//    return iterable(() -> {
//      if (it.hasNext())
//        return Optional.of(it.next());
//      else
//        return Optional.empty();
//    });
//  }

  /**
   * Creates an iterable with words which are read in from a reader.
   *
   * @param reader the reader to read in the words
   * @return the iterable with the words
   * @throws IOException thrown if read fails
   */
  public static Iterable<String> words(Reader reader) throws IOException {
    return iterable(() -> {
      return readNextWord(reader);
    });
  }

  private static Optional<String> readNextWord(Reader reader) {
    try {
      int x = reader.read();
      while (x != -1 && !Character.isLetterOrDigit(x)) {
        x = reader.read();
      }
      if (x == -1)
        return Optional.empty();
      StringBuilder b = new StringBuilder();
      while (x != -1 && Character.isLetterOrDigit(x)) {
        b.append((char) x);
        x = reader.read();
      }
      return Optional.of(b.toString());
    } catch (IOException e) {
      return Optional.empty();
    }
  }

  /**
   * Creates an iterable with provides the nodes starting from the start node and using
   * the successor function in a breadth-first manner.
   *
   * @param <N> the type of the nodes
   * @param start the start node
   * @param successorFn the successor function
   * @return the iterable with the nodes
   */
  public static <N> Iterable<N> breadthFirstIterable(N start, Function<N, List<N>> successorFn) {
    final Set<N> visited = new HashSet<>();
    final Queue<N> queue = new ArrayDeque<N>();
    queue.add(start);
    return iterable(() -> {
      if (queue.isEmpty())
        return Optional.empty();
      else {
        N node = queue.remove();
        visited.add(node);
        List<N> succs = successorFn.apply(node);
        succs.removeAll(visited);
        for (N n : succs) {
          if (!queue.contains(n))
            queue.add(n);
        }
        return Optional.of(node);
      }
    });
  }

  /**
   * Creates an iterator for the nodes starting from the start node and using
   * the successor function in a breadth-first manner.
   *
   * @param <N> the type of the nodes
   * @param start the start node
   * @param successorFn the successor function
   * @return the iterator with the nodes
   */
  public static <N> Iterator<N> breadthFirstIterator(N start, Function<N, List<N>> successorFn) {
    final Set<N> visited = new HashSet<>();
    final Queue<N> queue = new ArrayDeque<N>();
    queue.add(start);
    return iterator(() -> {
      if (queue.isEmpty())
        return Optional.empty();
      else {
        N node = queue.remove();
        visited.add(node);
        List<N> succs = successorFn.apply(node);
        succs.removeAll(visited);
        queue.addAll(succs);
        return Optional.of(node);
      }
    });
  }

  /**
   * A lazy iterator using a supplier function for generating the elements lazily.
   *
   * @param <T> the type of the elements
   */
  private static class LazyIterator<T> implements Iterator<T> {

    /**
     * The supplier function;
     * supplies an optional value with an empty optional
     * if no more values exist.
     */
    private final Supplier<Optional<T>> supplier;

    /**
     * An optional with the next value;
     * empty if no value exists;
     * null if a next value has to be generated
     */
    private Optional<T> optNext = null;

    /**
     * Private constructor.
     * @param supplier the supplier function
     */
    private LazyIterator(Supplier<Optional<T>> supplier) {
      this.supplier = supplier;
    }

    /**
     * Tests if a next value exists.
     * A next value exists if the optional value {@code optNext} is not empty.
     * Tests the {@code optNext} value is null and then uses the supplier
     * function to create a new optional value.
     *
     * @return {@code true} if there is a next value
     */
    @Override
    public boolean hasNext() {
      if (optNext == null)
        optNext = supplier.get();
      return optNext.isPresent();
    }

    /**
     * Returns the next value, if exists.
     * Throws a {@link NoSuchElementException} if no more element exists.
     *
     * @return the next value
     */
    @Override
    public T next() {
      if (!hasNext())
        throw new NoSuchElementException("No more");
      T result = optNext.get();
      optNext = null;
      return result;
    }

  }

}
