package at.jku.ssw.fp.sect05_2.workflow;

import java.util.function.BiFunction;
import java.util.function.Predicate;

import at.jku.ssw.fp.util.Pair;

/**
 * Functional interface for steps in a workflow.
 *
 * @author Herbert Praehofer
 *
 * @param <I> the type of the input of the step
 * @param <O> the type of the result of the step
 */
@FunctionalInterface
public interface Step<I, O> {

  /**
   * Executes this step.
   *
   * @param i the input of the step
   * @return the result of the step
   */
  O exec(I i);

  /**
   * Executes the step without an input (input is {@code null})
   * @return the result of the step
   */
  default O exec() {
    return this.exec(null);
  }

  /**
   * Combines this step with the step given as parameter.
   *
   * @param <R> the result of executing both steps
   * @param next the second step
   * @return the combined step
   */
  default <R> Step<I, R> then(Step<? super O, ? extends R> next) {
    return i -> next.exec(this.exec(i));
  }

  /**
   * Creates a combined step which executes either this step or the step given as parameter.
   * The decision which step to execute is based on a test of the input by the given predicate.
   *
   * @param that the other step
   * @param pred the predicate to test the input
   * @return the combined step
   */
  default Step<I, O> either(Step<? super I, ? extends O> that, Predicate<? super I> pred) {
    return i -> {
      if (pred.test(i))
        return this.exec(i);
      else
        return that.exec(i);
    };
  }

  /**
   * Creates a combined step which executes this step and the step given as a parameter and
   * combines the result of both in a {@link Pair}.
   *
   * @param <P> the result type of the second step
   * @param that the second step
   * @return the combined step
   */
  default <P> Step<I, Pair<O, P>> both(Step<? super I, ? extends P> that) {
    return i -> Pair.of(this.exec(i), that.exec(i));
  }

  /**
   * Creates a combined step which executes this step and the step given as a parameter and
   * combines the result by the given binary function.
   *
   * @param <P> the result type of the second step
   * @param <R> the type of the combined result
   * @param that the second step
   * @param combiner the combiner binary function
   * @return the combined step
   */
  default <P, R> Step<I, R> combine(Step<? super I, ? extends P> that, BiFunction<? super O, ? super P, R> combiner) {
    return i -> combiner.apply(this.exec(i), that.exec(i));
  }

  /**
   * Creates a step which executes the given step {@code n} times.
   *
   * @param <O> the type of the input and output (which are the same)
   * @param n the number of times to execute the step
   * @param step the step to repeat
   * @return the combined step with the repetitions
   */
  static <O> Step<O, O> repeat(int n, Step<? super O, ? extends O> step) {
    return o -> {
      O r = o;
      for (int i = 0; i < n; i++)
        r = step.exec(r);
      return r;
    };
  }
}
