package at.jku.ssw.fp.sect05_3.parser;

import static at.jku.ssw.fp.sect05_3.parser.Parser.alt;
import static at.jku.ssw.fp.sect05_3.parser.Parser.anyToken;
import static at.jku.ssw.fp.sect05_3.parser.Parser.token;

import at.jku.ssw.fp.coll.FList;
import at.jku.ssw.fp.expr.Expr;
import at.jku.ssw.fp.expr.Expr.And;
import at.jku.ssw.fp.expr.Expr.Lit;
import at.jku.ssw.fp.expr.Expr.Not;
import at.jku.ssw.fp.expr.Expr.Or;
import at.jku.ssw.fp.expr.Expr.Var;

/**
 * Class containing methods to create parser for parsing Boolean expressions.
 *
 * @author Herbert Praehofer
 *
 */
public class ExprParser {

  /**
   * Parser for Boolean literals.
   */
  public final Parser<Lit> litParser =
      alt(token("true").map(t -> new Lit(true)),
          token("false").map(t -> new Lit(false)));


  /**
   * Parser for variables.
   */
  Parser<Var> varParser =
      anyToken.filter(t -> ! t.equals("true") && ! t.equals("false")
                           &&  Character.isLetter(t.charAt(0)))
          .map(t -> new Var(t));


  /**
   * Parser for not-expressions.
   */
  public final Parser<Not> notParser =
      token("!")
        .seq(exprParser())
        .map(pair -> new Not(pair.snd));

  /**
   * Parser for or-expressions.
   */
  public final Parser<Or> orParser =
      token("||")
      .seq(exprParser())
      .seq(exprParser())
      .map(ppair -> new Or(ppair.fst.snd, ppair.snd));


  /**
   * Parser for and-expressions.
   */
  public final Parser<And> andParser =
      token("&&")
        .seq(exprParser())
        .seq(exprParser())
        .map(ppair -> Expr.mkAnd(ppair.fst.snd, ppair.snd));

  /**
   * Defected implementation of expression parser.
   * Using parser in an alternative results in endless loops.
   */
  public final Parser<Expr> exprParserDefected = alt(litParser, varParser, notParser, andParser, orParser);

  /** Variable to memoize the created expression parser. */
  public Parser<Expr> exprParser = null;

  /**
   * Creates an parser for Boolean expressions.
   * It uses the method for creating a parser from a list of supplier functions
   * for the alternatives.
   * Memoizes the parser once created.
   *
   * @return the parser for Boolean expressions.
   */
  public Parser<Expr> exprParser() {
     if (exprParser == null) {
      exprParser = alt(() -> litParser,
          () -> varParser,
          () -> notParser,
          () -> andParser,
          () -> orParser
      );

    }
    return exprParser;
  }

  /**
   * Applies the expression parser on a list of tokens.
   *
   * @param tokens the list of tokens
   * @return the parse result
   */
  public Result<Expr> parse(FList<String> tokens) {
    return exprParser().apply(tokens);
  }

  /**
   * Applies the expression parser on an input string.
   * From the input string a list of tokens is created.
   *
   * @param input the input string
   * @return the parse result
   */
  public Result<Expr> parse(String input) {
    return parse(FList.of(input.split(" ")));
  }

}
