package at.jku.ssw.fp.sect05_3.parser;

import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.function.Function;
import java.util.function.Predicate;
import java.util.function.Supplier;

import at.jku.ssw.fp.coll.FList;
import at.jku.ssw.fp.util.Pair;

/**
 * Functional interface for parsers.
 *
 * @author Herbert Praehofer
 *
 * @param <E> the type of parse result.
 */
@FunctionalInterface
public interface Parser<E> extends Function<FList<String>, Result<E>> {

  // Inherits Result<R> apply(FList<String> t);

  /**
   * Creates a new parser from this parser which is only successful
   * if the parse result fulfills the given predicate.
   *
   * @param pred the predicate to test the parse result
   * @return the new parser
   */
  default Parser<E> filter(Predicate<? super E> pred) {
    return input -> {
      Result<E> thisResult = this.apply(input);
      if (thisResult.failed())
        return thisResult;
      if (pred.test(thisResult.result))
        return thisResult;
      else
        return Result.failure(new NoSuchElementException("Filter failed"), input);

    };
  }

  /**
   * Creates a new parser from this parser where the parse result is mapped by
   * the given function.
   *
   * @param <R> the type of the parse result of the new parser
   * @param f the mapping function
   * @return the new parser
   */
  default <R> Parser<R> map(Function<? super E, ? extends R> f) {
    return input -> {
      Result<E> thisResult = this.apply(input);
      if (thisResult.successful())
        return Result.success(f.apply(thisResult.result), thisResult.rest);
      else
        return Result.failure(thisResult.excpt, thisResult.rest);
    };
  }

  /**
   * Creates a new parser which applies this parser and the second parser
   * and combined both results in a {@link Pair}.
   *
   * @param <S> the type of the parse result of the second parser
   * @param second the second parser
   * @return the combined parser
   */
  default <S> Parser<Pair<E, S>> seq(Parser<S> second) {
    return input -> {
      Result<E> thisRslt = this.apply(input);
      if (thisRslt.failed())
        return Result.failure(thisRslt.excpt, input);
      Result<S> sndRslt = second.apply(thisRslt.rest);
      if (sndRslt.failed())
        return Result.failure(sndRslt.excpt, input);
      return Result.success(Pair.of(thisRslt.result, sndRslt.result), sndRslt.rest);
    };
  }

  /**
   * Creates a parser which tries to apply one of the given alternative parsers.
   *
   * @param <R> the common type of the parse results of the alternative parsers.
   * @param alternatives the alternative parsers
   * @return the new parser
   */
  @SuppressWarnings("unchecked")
  @SafeVarargs
  static <R> Parser<R> alt(Parser<? extends R>... alternatives) {
    return input -> {
      Result<? extends R> result = null;
      for (Parser<? extends R> p : alternatives) {
        result = p.apply(input);
        if (result.successful())
          return (Result<R>) result;
      }
      return Result.failure(new ParseException("", input), input);
    };
  }

  /**
   * Creates a parser which tries to apply one of the alternative parsers.
   * The alternative parsers are created by the given supplier functions.
   *
   * @param <R> the common type of the parse results of the alternative parsers.
   * @param pSuppliers the supplier functions for the alternative parsers
   * @return the new parser
   */
  @SuppressWarnings("unchecked")
  @SafeVarargs
  static <R> Parser<R> alt(Supplier<Parser<? extends R>>... pSuppliers) {
    return input -> {
      Result<? extends R> result = null;
      for (Supplier<Parser<? extends R>> pSupplier : pSuppliers) {
        result = pSupplier.get().apply(input);
        if (result.successful())
          return (Result<R>) result;
      }
      return Result.failure(new ParseException("Parser alt failed", input), input);
    };
  }

  /**
   * Creates a parser which applies the given parser as long as it is successful.
   * The parse result is a list of the parse results of the multiple applications.
   *
   * @param <R> the type of the parse results of the given parser
   * @param parser the parser to apply multiple times
   * @return the parser applying the given parser multiple times
   */
  static <R> Parser<FList<R>> mult(Parser<R> parser) {
    return input -> {
      FList.Builder<R> b = FList.Builder.create();
      Result<? extends R> r = parser.apply(input);
      while (r.successful()) {
        b.add(r.result);
        r = parser.apply(r.rest);
      }
      return Result.success(b.build(), r.rest);
    };
  }

  /**
   * Creates a parser which optionally applies the given parser.
   * The result of the created parser is an optional.
   * If the parser is successful, then the parse result is returned
   * in an optional; otherwise the optional is empty and no token is consumed.
   * The parser is always successful but possibly with an empty result.
   *
   * @param <R> the type of the parse result
   * @param parser the parser to optionally apply
   * @return the optional parser
   */
  static <R> Parser<Optional<R>> opt(Parser<R> parser) {
    return input -> {
      Result<? extends R> r = parser.apply(input);
      if (r.successful())
        return Result.success(Optional.of(r.result), r.rest);
      else
        return Result.success(Optional.empty(), r.rest);
    };
  }

  /** A parser parsing any token. */
  Parser<String> anyToken = input -> input.size() != 0 ? Result.success(input.head(), input.tail())
      : Result.failure(new NoSuchElementException("Empty input"), input);

  /**
   * A parser parsing a token if it is equal to the given string.
   *
   * @param tkn the token to parse
   * @return the parser parsing a token which is equal to {@code tkn}.
   */
  static Parser<String> token(String tkn) {
    return anyToken.filter(t -> t.equals(tkn));
  }

}
