package at.jku.ssw.fp.sect05_3.parser;

import at.jku.ssw.fp.coll.FList;

/**
 * Class representing a result of a parser.
 * A result of a parser consists of the parse result and the rest of the input.
 * The parse result is either an real result or an exception. One of the two is not-null.
 *
 * @author Herbert Praehofer
 *
 * @param <T> the type of the parse result
 */
public class Result<T> {

  /**
   * Creates a successful result consisting of a parse result and the rest of the input.
   * The exception is {@code null}.
   *
   * @param <T> the type of the parse result
   * @param result the parse result
   * @param rest the rest of the input
   * @return the result object
   */
  public static <T> Result<T> success(T result, FList<String> rest) {
    return new Result<T>(result, rest);
  }

  /**
   * Creates a failed result consisting of an exception and the rest of the input.
   * The parse result is {@code null}.
   *
   * @param <T> the type of the parse result (used for type checking)
   * @param excpt the exception showing the parse failure
   * @param rest the rest of the input
   * @return the failed result object
   */
  public static <T> Result<T> failure(Exception excpt, FList<String> rest) {
    return new Result<T>(excpt, rest);
  }

  /** The parse result (maybe {@code null}).*/
  public final T result;

  /** The exception (maybe {@code null}).*/
  public final Exception excpt;

  /** The list of tokens not parsed yet. */
  public final FList<String> rest;

  /**
   * Private constructor creating a successful result object.
   *
   * @param result the parse result
   * @param rest the rest of input tokens
   */
  private Result(T result, FList<String> rest) {
    this.result = result;
    this.excpt = null;
    this.rest = rest;
  }

  /**
   * Private constructor creating a failed result object.
   *
   * @param excpt the exception showing the parse failure
   * @param rest the rest of input tokens
   */
  private Result(Exception excpt, FList<String> rest) {
    this.result = null;
    this.excpt = excpt;
    this.rest = rest;
  }

  /**
   * Returns if this object represents a successful result.
   *
   * @return {@code true} if this object represents a successful result
   */
  public boolean successful() {
    return result != null;
  }

  /**
   * Returns if this object represents a failed result.
   *
   * @return {@code true} if this object represents a failed result
   */
  public boolean failed() {
    return excpt != null;
  }

  @Override
  public String toString() {
    return "Result [result=" + result + ", exception=" + excpt + ", rest=" + rest.toString() + "]";
  }

}
