package at.jku.ssw.fp.sect05_4.dsl;

import static at.jku.ssw.fp.sect05_4.dsl.States.EMPTY;
import static at.jku.ssw.fp.sect05_4.dsl.States.FILLED;
import static at.jku.ssw.fp.sect05_4.dsl.States.FILLING;
import static at.jku.ssw.fp.sect05_4.dsl.States.STOPPED;

/** Enumeration for the states of the level controller */
enum States { EMPTY, FILLED, FILLING, STOPPED }

/**
 * State transition model for controlling the level of a tank.
 *
 * @author Herbert Praehofer
 *
 */
public class LevelControl extends StateModel<States> {

  /** Constants for full und empty level and constant inflow */
  static final double FULL_LEVEL = 100, EMPTY_LEVEL = 10,
            INFLOW = 10;

  /** Variables for the current level */
  private final DVar level;

  /** Variables for the current level and the inflow (which is a control output) */
  private final DVar inflow;

  /** Variable for the Boolean input signal to stop filling */
  private BVar stop;

  /**
   * Constructor for building the level control system.
   * The input and out variables are set.
   * Then the transition model is built using the DSL.
   *
   * @param _level the variable for the current level
   * @param _stop the variable for the stop signal
   * @param _inflow the variable for the output of the controller defining the inflow to the tank
   */
  public LevelControl(DVar _level, BVar _stop, DVar _inflow) {
    super(States.values(), FILLED);
    this.inflow = _inflow;
    this.stop = _stop;
    this.level = _level;

    // state transition model
    inState(EMPTY)
      .on(() -> true,
       () -> {
        inflow.set(INFLOW);
        nextState(FILLING);
       });
    inState(FILLING)
      .on(() -> stop.get(),
       () -> {
        inflow.set(0.0);
        nextState(STOPPED);
       })
      .on(() -> level.get() >= FULL_LEVEL,
       () -> {
        inflow.set(0.0);
        nextState(FILLED);
       });
    inState(FILLED)
      .on(() -> level.get() < EMPTY_LEVEL,
       () -> {
        nextState(EMPTY);
       });
    inState(STOPPED)
      .on(() -> ! stop.get(),
       () -> {
        inflow.set(INFLOW);
        nextState(FILLING);
       });
  }

  @Override
  public String toString() {
    return "LevelControl [level=" + level + ", inflow=" + inflow + ", stop=" + stop + "]";
  }

}

