package at.jku.ssw.fp.sect05_4.dsl;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.function.BooleanSupplier;

/**
 * Base class for implementing state transition models.
 * State transition models have a set of states and
 * state transitions.
 * Provides methods for implementing the DSL for state transitions models.
 *
 * @author Herbert Praehofer
 *
 * @param <S> the type of the states; must be an enumeration type
 */
public abstract class StateModel<S extends Enum<S>> implements Action {

  /** Map from state value to state objects*/
  private Map<S, State> states = new HashMap<>();

  /** The current state */
  protected State current;

  /**
   * Constructor initializing the set of states and the initial state.
   *
   * @param states the set of states
   * @param initial the initial state
   */
  protected StateModel(S[] states, S initial) {
    for (S s : states) {
      this.states.put(s, new State());
    }
    current = this.states.get(initial);
  }

  /**
   * Tests if the given state is the current state.
   *
   * @param s the state to test if it is the current state
   * @return {@code true} if the state is the current state
   */
  protected State inState(S s) {
    return states.get(s);
  }

  /**
   * Sets the current state to the given state.
   *
   * @param s the new current state
   */
  protected void nextState(S s) {
    current = states.get(s);
  }

  /**
   * Executes the model by executing the current state object.
   */
  @Override
  public void exec() {
    current.exec();
  }

  /**
   * Inner class for state objects.
   */
  protected static class State implements Action {

    /** List of transition objects for this state */
    final List<Transition> transitions = new ArrayList<>();

    /**
     * Executes the state object by looking for a executable transition and
     * executing the action of the first transition with a condition being fulfilled.
     */
    public void exec() {
      for (Transition tr : transitions) {
        if (tr.cond.getAsBoolean()) {
          tr.action.exec();
          break;
        }
      }
    }

    /**
     * Adds a transition which triggers based on the given condition.
     *
     * @param cond the transition condition
     * @param action the transition action
     * @return this state object
     */
    public State on(BooleanSupplier cond, Action action) {
      transitions.add(new Transition(cond, action));
      return this;
    }

    /**
     * Inner class representing transitions.
     */
    private static class Transition {

      /** The function object for the transition condition */
      private BooleanSupplier cond;

      /** The action of the transition */
      private Action action;

      /**
       * Constructor initializing condition and action.
       *
       * @param cond the function object for the transition condition
       * @param action the transition action
       */
      private Transition(BooleanSupplier cond, Action action) {
        this.cond = cond;
        this.action = action;
      }

    }

  }

}
