package at.jku.ssw.fp.sect06_2;

import java.util.function.BinaryOperator;

/**
 * Interface representing monoids.
 * A monoid consists of
 * <ul>
 *   <li> a set of values represented by the generic type {@code M},
 *   <li> a identity value provided by method {@code identity()},
 *   <li> a binary combiner function represented by method {@code combine},
 * </ul>
 * @author Herbert Praehofer
 *
 * @param <M> the type of values
 */
public interface Monoid<M> {

  /**
   * Returns the identity value.
   *
   * @return the identity value
   */
  M identity();

  /**
   * Combines two values of the monoid and returns a combined value.
   *
   * @param a the first value
   * @param b the second value
   * @return the combined value
   */
  M combine(M a, M b);

  /**
   * Constructs a monoid from an identity value and an binary operator.
   *
   * @param <M> the type of value
   * @param idty the identity value
   * @param combiner the combiner function
   * @return the monoid object
   */
  static <M> Monoid<M> of(M idty, BinaryOperator<M> combiner) {
    return new Monoid<M>() {
      @Override
      public M identity() {
        return idty;
      }

      @Override
      public M combine(M a, M b) {
        return combiner.apply(a, b);
      }
    };
  }
}
