package at.jku.ssw.fp.sect06_2;

import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.NoSuchElementException;
import java.util.Optional;
import java.util.Set;

/**
 * Class with fields and methods providing monoid objects.
 *
 * @author Herbert Praehofer
 *
 */
public class Monoids {

  /**
   * The monoid for double values, the plus operator and {@code 0.0} as identity value.
   */
  public static Monoid<Double> doublePlusMonoid = Monoid.of(0.0, (x, y) -> x + y);

  /**
   * The monoid for double values, the times operator and {@code 1.0} as identity value.
   */
  public static Monoid<Double> doubleMultMonoid = Monoid.of(1.0, (x, y) -> x * y);

  /**
   * The monoid for integer values, the plus operator and {@code 0} as identity value.
   */
  public static Monoid<Integer> intPlusMonoid = Monoid.of(0, (x, y) -> x + y);

  /**
   * The monoid for integer values, the times operator and {@code 1} as identity value.
   */
  public static Monoid<Integer> intTimesMonoid = Monoid.of(1, (x, y) -> x * y);

  /**
   * The monoid for string values, the concatenation operator and {@code ""} as identity value.
   */
  public static Monoid<String> stringMonoid = Monoid.of("", (a, b) -> a + b);

  /**
   * The monoid for Boolean values, the and operator and {@code true} as identity value.
   */
  public static Monoid<Boolean> booleanAndMonoid = Monoid.of(true, (x, y) -> x && y);

  /**
   * The monoid for Boolean values, the or operator and {@code false} as identity value.
   */
  public static Monoid<Boolean> booleanOrMonoid = Monoid.of(false, (x, y) -> x || y);

  /**
   * Creates a monoid object for power sets for a generic value set {@code A},
   * the union operator and the empty set as identity value.
   *
   * @param <A> the type of values for the sets
   * @return the monoid object
   */
  public static <A> Monoid<Set<A>> setMonoid() {
    return Monoid.of(Collections.emptySet(), (Set<A> a, Set<A> b) -> {
      Set<A> u = new HashSet<A>(a);
      u.addAll(b);
      return u;
    });
  }

  /**
   * Creates a monoid object for list for a generic value set {@code A},
   * the list concatenation  operator and the empty list as identity value.
   *
   * @param <A> the type of values for the list elements
   * @return the monoid object
   */

  public static <A> Monoid<List<A>> listMonoid() {
    return Monoid.of(Collections.emptyList(), (List<A> a, List<A> b) -> {
      List<A> c = new ArrayList<A>(a);
      c.addAll(b);
      return c;
    });
  }

  /**
   * Creates a monoid for optional values where two values of non-empty
   * optionals are combined by the given monoid.
   *
   * @param <E> the values contained in the optional
   * @param eMonoid the monoid for the contained values
   * @return the monoid for the optional values
   */
  public static <E> Monoid<Optional<E>> optionalMonoid(Monoid<E> eMonoid) {
    return Monoid.of(Optional.empty(), (Optional<E> a, Optional<E> b) -> {
      if (a.isPresent() && b.isPresent())
        return Optional.of(eMonoid.combine(a.get(), b.get()));
      else if (a.isPresent())
        return a;
      else if (b.isPresent())
        return b;
      else
        return Optional.empty();
    });
  }

  /**
   * Creates a monoid for maps from a key to value where values are combined by the
   * given monoid. The monoid for the values is used to combine values when in both maps
   * entries for the same key exist.
   *
   * @param <K> the type of the keys  in the map
   * @param <V> the type of the values in the map
   * @param vMonoid the monoid for combining values
   * @return the monoid for combining maps
   */
  public static <K, V> Monoid<Map<K, V>> mapMonoid(Monoid<V> vMonoid) {
    return Monoid.of(Map.of(), (Map<K, V> a, Map<K, V> b) -> {
      Map<K, V> result = new HashMap<K, V>(a);
      for (Map.Entry<K, V> entry : b.entrySet()) {
        V aVal = result.get(entry.getKey());
        if (aVal == null) {
          result.put(entry.getKey(), entry.getValue());
        } else {
          result.put(entry.getKey(), vMonoid.combine(entry.getValue(), aVal));
        }
      }
      return result;
    });
  }

  /**
   * Returns a default monoid based on the type of a value.
   *
   * @param <M> the type of values
   * @param m a single value
   * @return a monoid for the value
   */
  @SuppressWarnings("unchecked")
  public static <M> Monoid<M> implicitMonoid(M m) {
    if (m instanceof String)
      return (Monoid<M>) Monoids.stringMonoid;
    if (m instanceof Integer)
      return (Monoid<M>) Monoids.intPlusMonoid;
    if (m instanceof Double)
      return (Monoid<M>) Monoids.doublePlusMonoid;
    if (m instanceof Set)
      return (Monoid<M>) Monoids.setMonoid();
    if (m instanceof List)
      return (Monoid<M>) Monoids.listMonoid();
    else
      throw new NoSuchElementException("Not supported monoid for " + m.getClass().getName());
  }

}
