package at.jku.ssw.fp.sect06_2;

import static at.jku.ssw.fp.sect06_2.Monoids.intPlusMonoid;
import static at.jku.ssw.fp.sect06_2.Monoids.listMonoid;
import static at.jku.ssw.fp.sect06_2.Monoids.setMonoid;

import java.util.List;
import java.util.Set;
import java.util.function.Function;
import java.util.function.ToIntFunction;

import at.jku.ssw.fp.coll.Tree;

/**
 * Interface for reducible structures.
 *
 * @author Herbert Praehofer
 *
 * @param <E> the type of values to reduce
 */
public interface Reducible<E> {

  /**
   * Creates a reducible object for an iterable object.
   *
   * @param <E> the type of values
   * @param iterable the iterable object
   * @return the reducible object
   */
  static <E> Reducible<E> from(Iterable<E> iterable) {
    return new Reducible<E>() {

      @Override
      public <R> R reduceMap(Function<? super E, ? extends R> mapper, Monoid<R> monoid) {
        R result = monoid.identity();
        for (E elem : iterable)
          result = monoid.combine(result, mapper.apply(elem));
        return result;
      }

    };
  }

  /**
   * Creates a reducible object for an tree object.
   *
   * @param <E> the type of values
   * @param tree the tree object
   * @return the reducible object
   */
  public static <E> Reducible<E> from(final Tree<E> tree) {
    return new Reducible<E>() {

      @Override
      public <R> R reduceMap(Function<? super E, ? extends R> mapper, Monoid<R> monoid) {
        if (tree.isEmpty()) {
          return monoid.identity();
        } else {
          R rValue = mapper.apply(tree.value());
          R rLeft = Reducible.from(tree.left()).reduceMap(mapper, monoid);
          R rRight = Reducible.from(tree.right()).reduceMap(mapper, monoid);
          return monoid.combine(monoid.combine(rValue, rLeft), rRight);
        }
      }
    };
  }

  /**
   * First maps the elements and then reduces the mapped elements based on
   * the given monoid.
   *
   * @param <R> the type of the mapped values
   * @param mapper the mapping function
   * @param monoid the monoid for reduction
   * @return the reduced value
   */
  <R> R reduceMap(Function<? super E, ? extends R> mapper, Monoid<R> monoid);

  /**
   * Reduces the values based on the given monoid.
   * Uses {@code reduceMap} with an identity mapping.
   *
   * @param monoid the monoid for reduction
   * @return the reduced value
   */
  default E reduce(Monoid<E> monoid) {
    return reduceMap(Function.identity(), monoid);
  }

  /**
   * Creates a set of the elements.
   * The set is created by a reduction using the set monoid.
   *
   * @return the set of elements
   */
  public default Set<E> toSet() {
    return reduceMap(e -> Set.of(e), setMonoid());
  }

  /**
   * Creates a list of the elements.
   * The list is created by a reduction using the list monoid.
   *
   * @return the list of elements
   */
  default List<E> toList() {
    return reduceMap(e -> List.of(e), listMonoid());
  }

  /**
   * Counts the elements.
   * Maps each element to value 1 and does a reduction
   * using the monoid for integer values with the plus operator.
   *
   * @return the number of elements
   */
  public default int count() {
    return reduceMap(e -> 1, intPlusMonoid);
  }

  /**
   * Builds the sum of the mapped integer values.
   * First maps the elements to an integer value using the given
   * mapping function and then reduces the integer values
   * using the monoid for integer values with the plus operator.
   *
   * @param toIntFn the mapping function
   * @return the sum of mapped values
   */
  public default int sum(ToIntFunction<? super E> toIntFn) {
    return reduceMap(e -> toIntFn.applyAsInt(e), intPlusMonoid);
  }

  /**
   * Computes the minimum of mapped values.
   * First maps the elements to a comparable element and then does a reduction
   * using a monoid for the minimum value.
   *
   * @param <C> the type of mapped values
   * @param max the identity value for the monoid
   * @param toComparableFn the mapping function for mapping the elements to a comparable element
   * @return the minimum value
   */
  default <C extends Comparable<C>> C minimum(C max, Function<? super E, ? extends C> toComparableFn) {
    return reduceMap(toComparableFn, Monoid.of(max, (c1, c2) -> c1.compareTo(c2) < 0 ? c1 : c2));
  }

  /**
   * Computes the maximum of mapped values.
   * First maps the elements to a comparable element and then does a reduction
   * using a monoid for the maximum value.
   *
   * @param <C> the type of mapped values
   * @param min the identity value for the monoid
   * @param toComparableFn the mapping function for mapping the elements to a comparable element
   * @return the maximum value
   */
  default <C extends Comparable<C>> C maximum(C min, Function<? super E, ? extends C> toComparableFn) {
    return reduceMap(toComparableFn, Monoid.of(min, (c1, c2) -> c1.compareTo(c2) > 0 ? c1 : c2));
  }

}
