package at.jku.ssw.fp.sect06_3_4;

import java.util.function.Function;

/**
 * Class representing Logged monad.
 * Wraps a value and contains a log string.
 *
 * @author Herbert Praehofer
 *
 * @param <A> the type of values
 */
public class Logged<A> {

  /**
   * Creates a Logged monad with a wrapped value.
   *
   * @param <A>  the type of the value
   * @param a the wrapped value
   * @param log the log message
   * @return the monad with the wrapped value
   */
  public static <A> Logged<A> of(A a, String log) {
    return new Logged<A>(a, log);
  }

  /** The wrapped value */
  public final A value;

  /** The log string */
  public final String log;

  /**
   * Constructor wrapping value and setting log string.
   *
   * @param a the wrapped value
   * @param log the log string
   */
  private Logged(A a, String log) {
    this.value = a;
    this.log = log;
  }

  /**
   * Monad operator mapping value and concatenating log strings.
   *
   * @param <B>    the type of the wrapped result object.
   * @param mapper the mapping function
   * @return the wrapped result value
   */
  public <B> Logged<B> flatMap(Function<A, Logged<B>> mapper) {
    Logged<B> logged = mapper.apply(value);
    return Logged.of(logged.value, this.log + ";" + logged.log);
  }

  @Override
  public String toString() {
    return "Logged [value=" + value + ", log=" + log + "]";
  }

}

