package at.jku.ssw.fp.sect06_3_4;

import java.util.function.Function;

/**
 * Class representing Reader monad.
 * A reader monad reads a value from the context and
 * returns a value
 *
 * @author Herbert Praehofer
 *
 * @param <C> the type of the context
 * @param <A> the type of the value
 */
public class Reader<C, A> {

  /**
   * Creates a reader monad from a reader function.
   *
   * @param <C> the type of the context
   * @param <A> the type of the value
   * @param fn the reader function
   * @return the reader monad
   */
  public static <C, A> Reader<C, A> of(Function<C, A> fn) {
    return new Reader<>(fn);
  }

  /**
   * The reader function
   */
  private Function<C, A> fn;

  /**
   * Constructor setting the reader function.
   *
   * @param fn the reader function
   */
  private Reader(Function<C, A> fn) {
    this.fn = fn;
  }

  /**
   * Applies the reader function to a context.
   *
   * @param context the context
   * @return the read value
   */
  public A apply(C context) {
    return fn.apply(context);
  }

  /**
   * Monad operator.
   *
   * @param <B> the result type of the created reader monad.
   * @param mapper the mapping function
   * @return the resulting reader monad
   */
  public <B> Reader<C, B> flatMap(Function<? super A, Reader<C, ? extends B>> mapper) {
    return Reader.of(context -> mapper.apply(this.apply(context)).apply(context));
  }

}
