package at.jku.ssw.fp.sect07_4;

import java.util.ArrayDeque;
import java.util.ArrayList;
import java.util.Collections;
import java.util.Deque;
import java.util.List;
import java.util.Set;
import java.util.function.BiConsumer;
import java.util.function.BinaryOperator;
import java.util.function.Function;
import java.util.function.Supplier;
import java.util.stream.Collector;
import java.util.stream.Collectors;

/**
 * Collector grouping values with small differences.
 *
 * @author Herbert Praehofer
 *
 */
public class ConstantCollector
implements Collector<Double, Deque<List<Double>>, List<Double>> {

  /** Constant for the maximum difference between values which are regarded to be equal. */
  private static final double MAX_DIFF = 0.2;

  /**
   * A function that creates and returns a new queue of lists as mutable result container.
   *
   * @return a function which returns a new queue of lists
   */
  @Override
  public Supplier<Deque<List<Double>>> supplier() {
    return () -> {
      Deque<List<Double>> ql = new ArrayDeque<>();
      ql.add(new ArrayList<Double>());
      return ql;
    };
  }

  /**
   * A function that adds a value.
   * When the difference to the previous values is small, the value is added to the last list;
   * Otherwise a new list with this value is created and added to the queue of lists.
   *
   * @return a function which adds a value
   */
  @Override
  public BiConsumer<Deque<List<Double>>, Double> accumulator() {
    return (ql, msnt) -> {
      List<Double> l = ql.getLast();
      if (l.isEmpty() || smallDiff(l, msnt)) {
        l.add(msnt);
      } else {
        List<Double> nl = new ArrayList<>();
        nl.add(msnt);
        ql.add(nl);
      }
    };
  }

  /**
   * A function that accepts two partial results and merges them.
   *
   * @return a function which combines two partial results into a combined
   * result
   */
  @Override
  public BinaryOperator<Deque<List<Double>>> combiner() {
    return (q1, q2) -> {
      List<Double> q1Last = q1.getLast();
      List<Double> q2First = q2.getFirst();
      if (smallDiff(q1Last, q2First)) {
        q1Last.addAll(q2First);
        q2.removeFirst();
      }
      q1.addAll(q2);
      return q1;
    };
  }

  /**
   * Performs the final transformation. Each list in the queue of lists
   * is processed by computing the average of values. The result is a list of
   * average values.
   *
   * @return a function which transforms the intermediate result to the final
   * result
   */
  @Override
  public Function<Deque<List<Double>>, List<Double>> finisher() {
    return ql -> ql.stream()
           .map(l -> l.stream()
                .collect(Collectors.averagingDouble(mnt -> mnt)))
           .collect(Collectors.toList());
  }

  /**
   * Returns an empty set of characteristics.
   * @return empty set
   */
  @Override
  public Set<Characteristics> characteristics() {
    return Collections.emptySet();
  }

  private boolean smallDiff(List<Double> q, double msnt) {
    return Math.abs(avrg(q) - msnt) < MAX_DIFF;
  }

  private boolean smallDiff(List<Double> q1, List<Double> q2) {
    return Math.abs(avrg(q1) - avrg(q2)) < MAX_DIFF;
  }

  private double avrg(List<Double> l) {
    return l.stream().collect(Collectors.averagingDouble(mnt -> mnt));
  }

}
