package at.jku.ssw.fp.sect07_5.results;

import static at.jku.ssw.fp.util.When.is;
import static at.jku.ssw.fp.util.When.otherwise;
import static at.jku.ssw.fp.util.When.when;
import static java.util.stream.Collectors.counting;
import static java.util.stream.Collectors.groupingBy;
import static java.util.stream.Collectors.mapping;
import static java.util.stream.Collectors.partitioningBy;
import static java.util.stream.Collectors.toList;
import static java.util.stream.Collectors.toMap;

import java.io.IOException;
import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.Arrays;
import java.util.List;
import java.util.Map;
import java.util.stream.IntStream;
import java.util.stream.Stream;

/**
 * Demo for showing computation of assignment results using stream operations. 
 * 
 * @author Herbert Praehofer
 *
 */
public class Demo_Results {

  public static void main(String[] args) {

    try (Stream<String> stream = Files.lines(Paths.get("files/results.csv"))) {
      List<Results> resultList = 
          stream.skip(1)
                .map(l -> {
                  String[] elems = l.split(",");
                  Results r = new Results(elems[0], elems[1]);
                  for (int i = 2; i < elems.length; i++) {
                    try {
                      r.setScore(i - 1, Integer.parseInt(elems[i].trim()));
                    } catch (NumberFormatException e) {
                      r.setScore(i - 1, -1);
                    }
                  }
                  return r;
                })
                .collect(toList());

      System.out.format("%nResult list: %s %n", resultList);

      List<Integer> nSolved = resultList.stream().map(r -> r.countScores(s -> s >= 4)).collect(toList());
      System.out.format("%nNumbers solved: %s %n", nSolved);

      Map<String, Integer> nSolvedPerStnd = 
          resultList.stream().collect(
              toMap(r -> r.getName(), r -> r.countScores(s -> s >= 4)));

      System.out.format("%nNumbers solved: %s %n", nSolvedPerStnd);

      Map<Boolean, List<String>> sufficientSolved = 
          resultList.stream()
               .collect(partitioningBy(
                      r -> r.countScores(s -> s >= 4) >= 8, 
                      mapping(Results::getName, toList())));

      System.out.format("%nSufficient solved: %s %n", sufficientSolved);

      Map<String, Grade> grades = 
          resultList.stream()
                 .collect(toMap(r -> r.getName(), r -> computeGrade(r)));

      System.out.format("%nGrades: %s %n", grades);

      Map<Grade, Long> nStudentsWithGrade = 
          resultList.stream()
               .collect(groupingBy(r -> computeGrade(r), counting()));

      System.out.format("%nNumber for grade: %s %n", nStudentsWithGrade);

      Map<Boolean, Long> nPassedFailed = resultList.stream()
          .collect(partitioningBy(r -> computeGrade(r) != Grade.INSUFFICIENT, counting()));
      System.out.format("%nNumber passed/failed: %s %n", nPassedFailed);

      Map<String, Double> avrgScores = resultList.stream().collect(
          toMap(r -> r.getName(), r -> Arrays.stream(r.getScores()).filter(s -> s >= 0).average().getAsDouble()));
      System.out.format("%nAverage scores: %s %n", avrgScores);

      long nSolvedAss3 = resultList.stream().mapToInt(r -> r.getScore(3)).filter(s -> s >= 4).count();
      System.out.format("%nNumber solved assignments 3: %d %n", nSolvedAss3);

      List<Long> nSolvedPerAss = 
          IntStream
            .rangeClosed(1, 10)
            .mapToObj(i ->  resultList.stream()
                                      .mapToInt(r -> r.getScore(i))
                                      .filter(s -> s >= 4)
                                      .count())
            .collect(toList());

      System.out.format("%nNumber solved: %s %n", nSolvedPerAss);

      List<Long> nMaxPoints = IntStream.rangeClosed(1, 10).mapToObj(i -> {
        return resultList.stream().mapToInt(r -> r.getScore(i)).filter(s -> s == 10).count();
      }).collect(toList());
      System.out.format("%nNumber max points: %s %n", nMaxPoints);
      double avrgAss3 = resultList.stream().mapToInt(r -> r.getScore(3)).filter(s -> s >= 0).average().orElse(0);
      System.out.format("%nAverage points for assignments 3: %f %n", avrgAss3);

      List<Double> avrgPointsPerAss = 
          IntStream.rangeClosed(1, 10)
            .mapToObj(i -> resultList.stream()
                         .mapToInt(r -> r.getScore(i))
                         .filter(s -> s >= 0) 
                         .average()
                         .orElse(0.0) )
            .collect(toList()); 

      System.out.format("%nAverage points: %s %n", avrgPointsPerAss);

    } catch (IOException e) {
      e.printStackTrace();
    }
  }

  private static Grade computeGrade(Results r) {
    long n = r.countScores(s -> s >= 4); 
    if (n < 8) return Grade.INSUFFICIENT; 
    double avrg = 
      Arrays.stream(r.getScores())
            .sorted()
            .skip(2)
            .average()
            .getAsDouble();  
    return 
      when(avrg, 
        is(a -> a < 5.0, a -> Grade.INSUFFICIENT), 
        is(a -> a < 6.5, a -> Grade.SUFFICIENT), 
        is(a -> a < 8.0, a -> Grade.SATISFACTORY), 
        is(a -> a < 9.0, a -> Grade.GOOD), 
        otherwise(a -> Grade.EXCELLENT) ); 
  }


}
