package at.jku.ssw.fp.sect08_2;

import static at.jku.ssw.fp.sect08_2.Constants.logStream;

import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.PrintStream;
import java.util.ArrayList;
import java.util.Spliterator;
import java.util.concurrent.ExecutionException;
import java.util.concurrent.ForkJoinPool;
import java.util.concurrent.RecursiveTask;
import java.util.function.BinaryOperator;

/** 
 * Task for simulating the reduce operation of parallel streams 
 * including log information. 
 * 
 * @author Herbert Praehofer
 *
 * @param <V> the type of values processed
 */
@SuppressWarnings("serial")
class ReduceTask<V> extends RecursiveTask<V> {

  /** Threshold for splitting */ 
  private static final int THRESHOLD = 7; 
  
  /** The spliterator for this task */
  private final Spliterator<V> spltr; 
  
  /** The identity value for the reduction */
  private final V identity; 
  
  /** The binary operator for the reduction */
  private final BinaryOperator<V> accu, comb; 

  /** 
   * Constructor initializing the task. 
   * 
   * @param spltr the spliterator 
   * @param identity the identity value for the reduction 
   * @param accu the accumulator function for the sequential reduction 
   * @param comb the combiner function for the reduction 
   */
  public ReduceTask(Spliterator<V> spltr, V identity, BinaryOperator<V> accu, BinaryOperator<V> comb) {
    super();
    this.spltr = spltr;
    this.identity = identity;
    this.accu = accu;
    this.comb = comb;
  }

  /** The reduction value */
  V result;

  /**
   * Executes this task. Either splits or does a sequential reduction. 
   * 
   * @return the reduced value 
   */
  @Override
  protected V compute() {

    // accumulate sequentially
    long est = spltr.estimateSize();
    if (est <= THRESHOLD) {
      result = identity;
      spltr.forEachRemaining((V x) -> result = accu.apply(result, x));
      System.out.format("%nAccumulating: %s result %s %d", ((TracedSpliterator<V>) this.spltr).name, result,
          ((String) result).length());
      return result;
    }

    // split and fork
    ArrayList<ReduceTask<V>> subTasks = new ArrayList<ReduceTask<V>>();
    est = spltr.estimateSize();
    while (spltr.estimateSize() > THRESHOLD) {
      Spliterator<V> split = spltr.trySplit();
      ReduceTask<V> subTask = new ReduceTask<V>(split, identity, accu, comb);
      subTasks.add(subTask);
      System.out.format("%nForking: %s with size %d ", ((TracedSpliterator<V>) split).name, split.estimateSize());
      subTask.fork();
      est = spltr.estimateSize();
    }

    // accumulate rest of elements of this splitertor
    result = identity;
    spltr.forEachRemaining((V x) -> result = accu.apply(result, x));
    System.out.format("%nAccumulating: %s result %s %d", ((TracedSpliterator<V>) this.spltr).name, result,
        ((String) result).length());

    // join and combine

    for (ReduceTask<V> subTask : subTasks) {
      subTask.join();
      try {
        V pr = subTask.get();
        System.out.format("%nJoined: %s result %s %d", ((TracedSpliterator<V>) subTask.spltr).name, pr,
            ((String) pr).length());
        result = comb.apply(result, pr);
      } catch (InterruptedException | ExecutionException e) {
        e.printStackTrace(System.err);
      }
    }

    // return result
    return result;
  }

}

/**
 * Main demo class showing the reduction by the {@link ReduceTask}. 
 * 
 * @author Herbert Praehofer
 *
 */
public class Demo1_ReduceTask {

  public static void main(String[] args) throws FileNotFoundException {

    logStream = new PrintStream(new FileOutputStream("files/reduce" + Constants.N + ".log"));

    TracedList<String> listN = new TracedList<String>();
    for (int i = 0; i < Constants.N; i++) {
      listN.add(String.valueOf(i % 10));
    }
    System.out.println(listN);

    Spliterator<String> spliterator = listN.spliterator();

    ReduceTask<String> task = new ReduceTask<String>(spliterator, "", String::concat,
        String::concat);
    ForkJoinPool executor = new ForkJoinPool();

    String result = executor.invoke(task);
    System.out.println(result);

    // result = listN.parallelStream().reduce("", String::concat);
    System.out.println(result);

  }

}
