package at.jku.ssw.fp.sect08_2;

import static at.jku.ssw.fp.sect08_2.Constants.logStream;

import java.util.Spliterator;
import java.util.function.Consumer;

/**
 * Spliterator with tracing.
 * Wraps a the real spliterator.
 *
 * @author Herbert Praehofer
 *
 * @param <T> the type of the processed values
 */
class TracedSpliterator<T> implements Spliterator<T> {

  /** Reference to the real spliterator. */
  final Spliterator<T> realSpliterator;

  /** A name for the spliterator used for trace output. */
  final String name;

  /** Count for splits */
  int splits = 0;

  /**
   * Constructor setting the real spliterator and a name.
   *
   * @param realSpliterator the real spliterator
   * @param name the name
   */
  public TracedSpliterator(Spliterator<T> realSpliterator, String name) {
    super();
    this.name = name;
    this.realSpliterator = realSpliterator;
  }

  /**
   * Applies the consumer action on the next element.
   * Calls the real spliterator and additionally writes some trace information.
   *
   * @param action the consumer function
   * @return {@code true} if next element exists
   */
  @Override
  public boolean tryAdvance(Consumer<? super T> action) {
    logStream.format("%n%s: %s - tryAdvance: size = %d", Thread.currentThread().getName(), name, this.estimateSize());
    return realSpliterator.tryAdvance(action);
  }

  /**
   * Tries a split and returns a spliterator.
   * Calls the real spliterator and additionally writes some trace information.
   *
   * @return the split spliterator
   */
  @Override
  public Spliterator<T> trySplit() {
    Spliterator<T> realSplit = realSpliterator.trySplit();
    if (realSplit == null)
      return null;
    splits++;
    String splitName = name + splits + ".";
    logStream.format("%n%s: %s - trySplit: splitted %s, size = %d", Thread.currentThread().getName(), name, splitName,
        realSplit.estimateSize());
    return new TracedSpliterator<>(realSplit, splitName);
  }

  /**
   * Asks the real spliterator for the estimated size.
   *
   * @return the estimated size
   */
  @Override
  public long estimateSize() {
    long est = realSpliterator.estimateSize();
    // Constants.logStream.format("%n%s: %s - estimateSize = %d ",
    // Thread.currentThread().getName(), name, est);
    return est;
  }

  /**
   * Calls the real spliterator for returning the characteristics.
   *
   * @return the characteristics
   */
  @Override
  public int characteristics() {
    // Constants.logStream.format("%n%s: %s - characteristics",
    // Thread.currentThread().getName(), name);
    return realSpliterator.characteristics();
  }

}
