package at.jku.ssw.fp.sect10_1;

import static at.jku.ssw.fp.sect10.Constants.executor;

import java.time.LocalDate;
import java.util.concurrent.CompletableFuture;
import java.util.concurrent.atomic.AtomicInteger;

import at.jku.ssw.fp.airline.Airline;
import at.jku.ssw.fp.airline.Flight;
import io.reactivex.Observable;
import io.reactivex.Observer;
import io.reactivex.disposables.Disposable;
import io.reactivex.schedulers.Timed;

/**
 * Demo showing an observable from asynchronous results. 
 * 
 * @author Herbert Praehofer
 *
 */
public class Demo3_AsyncFlightObservable {

  public static void main(String[] args) {
    
    LocalDate nextWeek =  LocalDate.now().plusDays(7); 

    Airline[] airlines = Airline.AIRLINES;

    long start = System.currentTimeMillis();
      
    Observable<Flight> flightObservable = 
        Observable.create(emitter -> {
          AtomicInteger nFtrs = 
              new AtomicInteger(airlines.length);
          for (Airline airline : airlines) {
            CompletableFuture.supplyAsync(() -> { 
                  return airline.findFlight("VIE", "JKF", nextWeek);
                }, executor)
                .whenComplete((flight, excpt) -> {
                  synchronized (emitter) { 
                    if (flight != null) 
                      emitter.onNext(flight);
                    if (nFtrs.decrementAndGet() == 0)  
                        emitter.onComplete();
                    } 

                }); 
          }
       });
    
    flightObservable.subscribe(new Observer<Flight>() {

      @Override
      public void onSubscribe(Disposable d) {
        System.out.println("ONSUBSCRIBE");
      }

      @Override
      public void onNext(Flight flight) {
        System.out.println("onNext: " + Thread.currentThread().getName()); 
        System.out.println("ONNEXT: " + flight.toString());
      }

      @Override
      public void onError(Throwable excpt) {
        System.out.println("onError: " + Thread.currentThread().getName()); 
        System.out.println("ONERROR: " + "Exception in " + excpt.toString());
      }

      @Override
      public void onComplete() {
        System.out.println("onComplete: " + Thread.currentThread().getName()); 
        System.out.println("COMPLETED");
      }
      
    });
    
    // with time stamp 

    Observable<Timed<Flight>> timedFlightObservable = 
        flightObservable.timestamp();
 
    
    timedFlightObservable.subscribe(new Observer<Timed<Flight>>() {

      @Override
      public void onSubscribe(Disposable d) {
      }

      @Override
      public void onNext(Timed<Flight> timedFlight) {
        System.out.println("onNext: " + Thread.currentThread().getName()); 
        System.out.println(timedFlight.time() - start + " ms");
        System.out.println(timedFlight.value().toString());
      }

      @Override
      public void onError(Throwable excpt) {
        System.out.println("onError: " + Thread.currentThread().getName()); 
        System.out.println("Exception in " + excpt.toString());
      }

      @Override
      public void onComplete() {
        System.out.println("onComplete: " + Thread.currentThread().getName()); 
        System.out.println("COMPLETED");
      }
      
    });
    
    System.out.format("Execution time: %d", System.currentTimeMillis() - start);

  }

}
