package at.jku.ssw.fp.sect10_1;

import java.awt.Dimension;
import java.awt.Graphics;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.event.MouseAdapter;
import java.awt.event.MouseEvent;
import java.util.ArrayList;
import java.util.EventListener;
import java.util.EventObject;
import java.util.HashSet;
import java.util.List;
import java.util.Set;

import javax.swing.JFrame;
import javax.swing.JPanel;
import javax.swing.WindowConstants;

import io.reactivex.Observable;
import io.reactivex.Observer;
import io.reactivex.disposables.Disposable;

/**
 * Demos showing observables from mouse clicks. 
 * 
 * @author Herbert Praehofer
 *
 */
public class Demo4_ClickObservable {

  private static JFrame frame; 
  
  public static void main(String[] args) {
    
    frame = new JFrame("Observable Demo"); 
    frame.setDefaultCloseOperation(WindowConstants.EXIT_ON_CLOSE);
    ObservablePanel panel = new ObservablePanel(new PointsModel()); 
    frame.setContentPane(panel);
    frame.pack(); 
    frame.setLocation(200, 200);
    frame.setVisible(true);
    
  }
  
  static class PointsModel {
    private Set<Point> points = new HashSet<>(); 
    private List<PointListener> pointListeners = new ArrayList<>();
    
    public void addPoint(Point point) {
      points.add(point); 
      PointEvent pe = new PointEvent(this, point);
      for (PointListener l : pointListeners) {
        l.pointAdded(pe);
      }
    }
    
    public void addPointListener(PointListener l) {
      pointListeners.add(l); 
    }
    public void removePointListener(PointListener l) {
      pointListeners.remove(l); 
    }
  }
  
  static interface PointListener extends EventListener {
    void pointAdded(PointEvent pe); 
  }
  
  @SuppressWarnings("serial")
  static class PointEvent extends EventObject {

    public final Point point; 
    public PointEvent(Object source, Point point) {
      super(source);
      this.point = point; 
    }
    
  }
  
  @SuppressWarnings("serial")
  static class ObservablePanel extends JPanel {
    
    final Observable<Point> clickObservable; 
    final PointsModel model; 
    ObservablePanel(PointsModel model) {
      this.model = model; 
      model.addPointListener(new PointListener() {

        @Override
        public void pointAdded(PointEvent pe) {
          repaint(); 
        }
        
      });
      ObservablePanel panel = this; 
      this.setPreferredSize(new Dimension(300, 300));
      
      clickObservable = Observable.create(
          emitter -> {
            panel.addMouseListener(
              new MouseAdapter() {
                @Override
                public void mouseClicked(MouseEvent e) {
                  emitter.onNext(e.getPoint());
                }
              }); 
          });

      final Rectangle area = new Rectangle(10, 10, getWidth() - 20, getHeight() - 20);
      clickObservable.subscribe(p -> System.out.println(p)); 
      clickObservable
      .filter(point -> isWithinArea(point, area))
      .map(point -> snapToGrid(point))
      .subscribe(new Observer<Point>() {
  
          @Override
          public void onSubscribe(Disposable d) {}
  
          @Override
          public void onNext(Point p) {
            System.out.println(p);
            model.addPoint(p);
          }
  
          @Override
          public void onError(Throwable e) {}
  
          @Override
          public void onComplete() {}
      });
      
    }
    
    @Override
    protected void paintComponent(Graphics g) {
      super.paintComponent(g);
      for (Point p: model.points) {
        g.drawRect(p.x, p.y, 2, 2);
      }
    }

    private boolean isWithinArea(Point p, Rectangle rect) {
      return p.x >= rect.x && p.x <= rect.x + rect.width
          && p.y >= rect.y && p.y <= rect.y + rect.height; 
    }
    
    private Point snapToGrid(Point p) {
      return new Point((p.x + 5 ) / 10 * 10, (p.y + 5) / 10 * 10); 
    }
  }
  

}
