package at.jku.ssw.fp.sect10_7;

import java.nio.file.Files;
import java.nio.file.Paths;
import java.util.Scanner;
import java.util.concurrent.CountDownLatch;
import java.util.concurrent.TimeUnit;
import java.util.stream.Stream;

import org.reactivestreams.Subscriber;
import org.reactivestreams.Subscription;

import io.reactivex.BackpressureStrategy;
import io.reactivex.Flowable;
import io.reactivex.schedulers.Schedulers;

/**
 * Demo program implementing a streaming of a text with backpressure. 
 * Uses explicit requests from consumer to control the flow. 
 * 
 * @author Herbert Praehofer
 */
public class Demo4_StreamingSongWithRequests {
	
	private static final int DELAY = 20;
	private static CountDownLatch teriminateLatch; 
	
	public static void main(String[] args) throws InterruptedException {
		
		try (Scanner scanner = new Scanner(System.in)) {
			System.out.print("Bitte Songdatei eingeben (adele | cohen | sinatra): ");
			String input = scanner.nextLine().trim(); 
			String fileName = "songs\\" + input + ".txt"; 
			if (Files.notExists(Paths.get(fileName))) {
				System.out.println(String.format("File %s existiert nicht ", fileName));
				return; 
			} 
			
			Flowable<Character> charFlowable = 
					Flowable.<Character>create(emitter -> {
						try (Stream<String> lines = Files.lines(Paths.get(fileName))) {
							lines
								.map(l -> l + "\n")
								.flatMapToInt(l -> l.chars())
								.forEach(c -> emitter.onNext((char)c));
							emitter.onComplete();
						} catch (Exception e) {
							emitter.onError(e);
						}
					}, BackpressureStrategy.BUFFER)
					.subscribeOn(Schedulers.io()); 
			
			
			teriminateLatch = new CountDownLatch(1); 
			Flowable.interval(0, DELAY, TimeUnit.MILLISECONDS)
	  		  .zipWith(charFlowable, (i, c) -> c)
	  		  .observeOn(Schedulers.computation())
	  		  .subscribe(
	  		    new Subscriber<Character>() {
	  		      private Subscription subscription; 
	  		      private int count; 
	  
	  		      @Override
	  		      public void onSubscribe(Subscription subscription) {
	  		        this.subscription = subscription; 
	  		        subscription.request(20);
	  		        count = 20; 
	  		      }
	  
	  		      @Override
	  		      public void onNext(Character c) {
	  		        System.out.print(c);
	  		        try {Thread.sleep(DELAY);} catch (Exception e) {} 
	  		        count--; 
	  		        if (count == 10) {
	  		          subscription.request(10);
	  		          count = count + 10; 
	  		        }
	  		      }
	  
	  					@Override
	  					public void onError(Throwable t) {
	  						System.out.println("ERROR: " + t); 
	  						teriminateLatch.countDown();
	  					}
	  
	  					@Override
	  					public void onComplete() {
	  						System.out.println("END");
	  						teriminateLatch.countDown();
	  					}
	  
	  				});
			
			teriminateLatch.await(); 
		}
	}
	
}
