package at.jku.ssw.fp.sect11_3.check4j;

import java.util.function.Function;
import java.util.function.Predicate;
import java.util.stream.Stream;

/**
 * Functional interface for random value generators.
 * Random value generators form a monad.
 *
 * @author Herbert Praehofer
 *
 * @param <A> the type of the generated values
 */
@FunctionalInterface
public interface Gen<A> {

  /**
   * Returns a new random value.
   *
   * @return the new random value
   */
  A sample();

  /**
   * Returns a stream of random values.
   * @return the stream of random values
   */
  default Stream<A> stream() {
    return Stream.generate(() -> sample());
  }

  /**
   * Creates a generator returning the provided value.
   *
   * @param <A> the type of the returned value
   * @param a the constant value to return
   * @return the constant value
   */
  static <A> Gen<A> of(A a) {
    return () -> a;
  }

  /**
   * Creates a generator which generates values which are created by
   * mapping the values from this generator using the mapping function.
   *
   * @param <B> the type of mapped value
   * @param fn the mapping function
   * @return the generator for the mapped values
   */
  default <B> Gen<B> map(Function<? super A, ? extends B> fn) {
    return () -> fn.apply(this.sample());
  }

  /**
   * Creates a generator which generates values which are created by
   * mapping the values from this generator using the mapping function.
   *
   * @param <B> the type of mapped value
   * @param fn the mapping function
   * @return the generator for the mapped values
   */
  default <B> Gen<B> flatMap(
                 Function<? super A, Gen<? extends B>> fn) {
    return () -> fn.apply(this.sample()).sample();
  }

  /**
   * Filters the value of this generator. If the value
   * generated by this generator is not accepted by the
   * predicate, a new value is generated by this generator.
   *
   * @param pred the testing function
   * @return the generator only generating values accepted by the predicate
   */
  default Gen<A> filter(Predicate<? super A> pred) {
    return () -> {
        final A v = this.sample();
        if (pred.test(v)) return v;
        else return this.sample();
      };
  }

}

