package at.jku.ssw.fp.sect11_3.check4j;

import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Random;

import at.jku.ssw.fp.coll.FList;
import at.jku.ssw.fp.util.Pair;
import at.jku.ssw.fp.util.Tuple2;
import at.jku.ssw.fp.util.Tuple3;

/**
 * Class with variables and methods providing various generators.
 *
 * @author Herbert Praehofer
 *
 */
public class Gens {

  /** A random number generator used for creating random values */
  private static final Random RAND = new Random();

  /**
   * A generator for double values (in full range of double)
   */
  public static Gen<Double> rands = () -> RAND.nextDouble();

  /**
   * A generator for integer values (in full range of integer)
   */
  public static Gen<Integer> intRands = () -> RAND.nextInt();

  /**
   * Creates a generator for double values between {@code from} and {@code to}
   * @param from the lower value
   * @param to the upper value
   * @return the generator for integer values between {@code from} and {@code to}
   */
  public static Gen<Double> randsFromTo(final double from, final double to) {
    double w = to - from;
    return rands.map(r -> from + w * r);
  }

  /**
   * Creates a generator for percentage values.
   */
  public static Gen<Double> percents = randsFromTo(0.0, 100.0);

  /**
   * Creates a generator for int values between {@code from} and {@code to}.
   *
   * @param from the lower value (inclusive)
   * @param to the upper value (exclusive)
   * @return the generator for integer values between {@code from} and {@code to}
   */
  public static Gen<Integer> intsFromTo(final int from, final int to) {
    final int w = to - from;
    return rands.map(r -> from + (int) (w * r));
  }

  /**
   * Creates a generator for int values between {@code 0} and {@code to}.
   *
   * @param to the upper limit (exclusive)
   * @return the generator for integer values between {@code 0} and {@code to}
   */

  public static Gen<Integer> intsTo(int to) {
    return rands.map(r -> (int) (to * r));
  }

  /**
   * A generator of positive int values
   */
  public static Gen<Integer> posInts = intsTo(Integer.MAX_VALUE);

  /**
   * A generator for in values int values in the full range of int.
   */
  public static Gen<Integer> allInts = intsFromTo(Integer.MIN_VALUE, Integer.MAX_VALUE);

  /**
   * A generator for Boolean value with equal probability.
   */
  public static Gen<Boolean> bools = rands.map(r -> r < 0.5);

  /**
   * Creates  generator for Boolean values where {@code true} values are
   * created with the given probability.
   *
   * @param prob the probability for {@code true}
   * @return the generator of Boolean values
   */
  public static Gen<Boolean> withProb(final double prob) {
    return rands.map(r -> r < prob);
  }

  /**
   * Creates a generator which creates values from the given
   * set of values with equal probability.
   *
   * @param <A> the type of value
   * @param as the values
   * @return the generator for the values
   */
  @SuppressWarnings("unchecked")
  public static <A> Gen<A> oneOf(final A... as) {
    assert as.length > 0;
    return intsFromTo(0, as.length).map(idx -> as[idx]);
  }

  /**
   * Creates a generator which creates values from the given
   * set of values with equal probability.
   *
   * @param <A> the type of value
   * @param as the list of values
   * @return the generator for the values
   */
  public static <A> Gen<A> oneOf(final List<A> as) {
    assert !as.isEmpty();
    return intsTo(as.size()).map(i -> as.get(i));
  }

  /**
   * Creates a generator which is capable to create a single value.
   * For every call to {@code sample} the same value is returned.
   *
   * @param <A> the type of the value
   * @param a the constant value
   * @return the generator creating the same value
   */
  public static <A> Gen<A> single(final A a) {
    return new Gen<A>() {
      @Override
      public A sample() {
        return a;
      }
    };
  }

  /** List of letters from 'a' to 'z' */
  private static List<Character> alphabet = new ArrayList<Character>();
  {
    for (char c = 'a'; c <= 'z'; c = (char) (c + 1)) {
      alphabet.add(c);
    }
  }

  /**
   * A generator for creating letters from 'a' to 'z' with same probability.
   */
  public static Gen<Character> letters = oneOf('a', 'b', 'c', 'd', 'e', 'f', 'g', 'h', 'i', 'j', 'k', 'l', 'm', 'n',
      'o', 'p', 'q', 'r', 's', 't', 'u', 'v', 'w', 'x', 'y', 'z');

  /**
   * Creates a generator for generating words from random letters of given maximum length.
   *
   * @param maxLength the maximum length
   * @return the generator for words
   */
  public static Gen<String> words(int maxLength) {
    final Gen<Integer> lengthGen = intsFromTo(1, maxLength + 1);
    return lengthGen.map(n -> {
      final StringBuffer b = new StringBuffer();
      for (int i = 0; i < n; i++) {
        b.append(letters.sample());
      }
      return b.toString();
    });
  }

  /**
   * Creates a generator for generating lists from given elements of given maximum length.
   *
   * @param <A> the element type
   * @param maxLength the maximum length
   * @param elems the elements
   * @return the generator for lists
   */
  @SuppressWarnings("unchecked")
  public static <A> Gen<List<A>> lists(int maxLength, A... elems) {
    final Gen<Integer> lengthGen = intsFromTo(1, maxLength + 1);
    final Gen<A> elemGen = oneOf(elems);
    return lengthGen.map(n -> {
      final List<A> l = new ArrayList<A>();
      for (int i = 0; i < lengthGen.sample(); i++) {
        l.add(elemGen.sample());
      }
      return Collections.unmodifiableList(l);
    });
  }

  /**
   * Creates a generator for generating lists of given maximum length.
   * The elements are generated by the given generator.
   *
   * @param <A> the element type
   * @param maxLength the maximum length
   * @param elemGen the generator for elements
   * @return the generator for lists
   */
  public static <A> Gen<FList<A>> lists(int maxLength, Gen<A> elemGen) {
    return intsFromTo(0, maxLength + 1).map(n -> {
      FList.Builder<A> builder = FList.Builder.create();
      elemGen.stream().limit(n).forEach(e -> builder.add(e));
      return builder.build();
    });
  }

  /**
   * A generator for list of integers from 0 to 9
   */
  public static Gen<FList<Integer>> intLists = lists(5, intsFromTo(0, 10));

  /**
   * A generator of pairs of values.
   * The values are generated by two given generators.
   *
   * @param <A> the type of the elements of the first generator
   * @param <B> the type of the elements of the second generator
   * @param aGen the first generator
   * @param bGen the second generator
   * @return the generator of pairs of values
   */
  public static <A, B> Gen<Pair<A, B>> pair(Gen<A> aGen, Gen<B> bGen) {
    return aGen.flatMap(a -> bGen.map(b -> Pair.of(a, b)));
  }

  /**
   * A generator of tuples of two values.
   * The values are generated by two given generators.
   *
   * @param <A> the type of the elements of the first generator
   * @param <B> the type of the elements of the second generator
   * @param aGen the first generator
   * @param bGen the second generator
   * @return the generator of tuple of two values
   */
  public static <A, B> Gen<Tuple2<A, B>> tuple2(Gen<A> aGen, Gen<B> bGen) {
    return aGen.flatMap(a -> bGen.map(b -> Tuple2.of(a, b)));
  }

  /**
   * A generator of tuples of three values.
   * The values are generated by two given generators.
   *
   * @param <A> the type of the elements of the first generator
   * @param <B> the type of the elements of the second generator
   * @param <C> the type of the elements of the third generator
   * @param aGen the first generator
   * @param bGen the second generator
   * @param cGen the third generator
   * @return the generator of tuple of three values
   */
  public static <A, B, C> Gen<Tuple3<A, B, C>> tuple3(Gen<A> aGen, Gen<B> bGen, Gen<C> cGen) {
    return aGen.flatMap(a -> bGen.flatMap(b -> cGen.map(c -> Tuple3.of(a, b, c))));
  }

  /**
   * A generator for prime numbers from {@code 2} to {@code 100}.
   */
  Gen<Integer> primes100 = intsFromTo(2, 100).filter(x -> isPrime(x));

  private static boolean isPrime(int n) {
    for (int i = 2; i <= n / 2; i++) {
      if (n % i == 0) {
        return false;
      }
    }
    return true;
  }

}
