package at.jku.ssw.fp.sect11_3.check4j;

import java.io.PrintStream;
import java.util.Optional;
import java.util.function.Predicate;
import java.util.stream.Collectors;
import java.util.stream.Stream;

/**
 * A property object in a property-based test.
 * <p>
 * It consists of a generator for generating test values,
 * a predicate for testing test values,
 * and a shrinker for shrinking a failed test value.
 * </p>
 * @author Herbert Praehofer
 *
 * @param <A> the type of the test values
 */
public class Property<A> {

  /** Constant for the number of generated test values. */
  public final int N = 100;

  /**
   * Creates a property object from a generator, predicate and shrinker.
   *
   * @param <A> the type of test values
   * @param generator the generator for generating test values
   * @param predicate the predicate for testing test values
   * @param shrinker the shrinker for shrinking a failed test value
   * @return the property object
   */
  public static <A> Property<A> forAll(Gen<A> generator, Predicate<? super A> predicate, Shrinker<A> shrinker) {
    return new Property<A>(generator, predicate, shrinker);
  }

  /**
   * Creates a property object from a generator and predicate.
   * No shrinker is used.
   *
   * @param <A> the type of test values
   * @param generator the generator for generating test values
   * @param predicate the predicate for testing test values
   * @return the property object
   */
  public static <A> Property<A> forAll(Gen<A> generator, Predicate<? super A> predicate) {
    return new Property<A>(generator, predicate);
  }

  /** The generator for generating test values. */
  private final Gen<A> generator;

  /** The predicate for testing test values. */
  private final Predicate<? super A> predicate;

  /** The optional shrinker for shrinking a failed test value. */
  private final Optional<Shrinker<A>> optShrinker;

  /**
   * Constructor setting generator and predicate.
   *
   * @param generator the generator for generating test values
   * @param predicate the predicate for testing test values
   */
  private Property(Gen<A> generator, Predicate<? super A> predicate) {
    this(generator, predicate, null);
  }

  /**
   * Constructor setting generator, predicate and shrinker.
   *
   * @param generator the generator for generating test values
   * @param predicate the predicate for testing test values
   * @param shrinker the shrinker for shrinking a failed test value
   */
  private Property(Gen<A> generator, Predicate<? super A> predicate, Shrinker<A> shrinker) {
    this.generator = generator;
    this.predicate = predicate;
    this.optShrinker = Optional.ofNullable(shrinker);
  }

  /**
   * Runs the property-based test represented by this property object.
   * <p>
   * Generates N test values using the generator and test them
   * by the predicate. When the first fails the test is aborted.
   *
   * Does not shrink a failed value.
   * </p>
   *
   * @return the result of the test run
   */
  public Result<A> check() {
    Result<A> result = Result.create();
    for (int i = 0; i < N; i++) {
      A a = generator.sample();
      if (predicate.test(a)) {
        result.addSuccess(a);
      } else {
        result.failWith(a);
        return result;
      }
    }
    return result;
  }

  /**
   * Runs the property-based test represented by this property object.
   * <p>
   * Generates N test values using the generator and test them
   * by the predicate.
   * When the first fails, the failed value is shrunk by the shrinker
   * and the test is aborted.
   * </p>
   *
   * @return the result of the test run
   */
  public Result<A> checkAndShrink() {
    Result<A> result = check();
    if (result.failed()) {
      Stream<A> shrinked = optShrinker.get().shrink(result.getFailure().get());
      A simplest = shrinked.peek(l -> System.out.println("Created: " + l)).filter(predicate.negate())
          .peek(l -> System.out.println("Filtered: " + l)).findFirst().orElse(result.getFailure().get());
      result.failWith(simplest);
    }
    return result;
  }

  /**
   * Runs the property-based test represented by this property object.
   * <p>
   * Generates N test values using the generator and test them
   * by the predicate.
   * When the first fails, the test is aborted.
   * If a shrinker is set, the failed value is shrunk.
   * Prints out the result of the test on the given printer stream.
   * </p>
   *
   * @param printer the printer stream
   */
  public void checkVerbose(PrintStream printer) {
    Result<A> result = check();
    if (result.failed()) {
      printer.format("%nFAILED with %s%n", result.getFailure().get().toString());
      if (optShrinker.isPresent()) {
        Stream<A> shrinked = optShrinker.get().shrink(result.getFailure().get());
        A simplest = shrinked.peek(l -> System.out.println("Created: " + l)).filter(predicate.negate())
            .peek(l -> System.out.println("Filtered: " + l)).findFirst().orElse(result.getFailure().get());
        printer.format("%nMINIMAL failing is %s", simplest.toString());
      }
    } else {
      printer.format("%nSUCCEEDED with %d test cases%n", result.getSuccesses().size());
      printer.println(result.getSuccesses().stream().map(Object::toString).collect(Collectors.joining("\n")));
    }
  }

  /**
   * Runs the property-based test represented by this property object.
   * Prints out the result of the test on standard output.
   */
  public void checkVerbose() {
    checkVerbose(System.out);
  }

}
