package at.jku.ssw.fp.sect11_3.check4j;

import java.util.ArrayList;
import java.util.List;
import java.util.Optional;

/**
 * Test result with the list of successful test values and an optional failed value.
 *
 * @author Herbert Praehofer
 *
 * @param <A> the type of test values
 */
public class Result<A> {

  /** List of successful test values */
  private List<A> successes = new ArrayList<A>();

  /** The optional failed test value */
  private Optional<A> failure = Optional.empty();

  /**
   * Creates an empty result object.
   *
   * @param <A> the type of the test values
   * @return an empty result object
   */
  public static <A> Result<A> create() {
    return new Result<A>();
  }

  /**
   * Tests if the results are all successful.
   *
   * @return {@code true} if all tests were successful
   */
  public boolean successful() {
    return !failed();
  }

  /**
   * Tests if there is failed test.
   *
   * @return {@code true} if a test failed
   */
  public boolean failed() {
    return failure.isPresent();
  }

  /**
   * Gets the list of successful test values.
   *
   * @return the list of successful test values
   */
  public List<A> getSuccesses() {
    return successes;
  }

  /**
   * Gets the failed test value in an optional; maybe empty.
   *
   * @return the optional with a  failed value or empty
   */
  public Optional<A> getFailure() {
    return failure;
  }

  /**
   * Adds a value to the list of successful test values.
   *
   * @param a the successful test value
   */
  public void addSuccess(A a) {
    successes.add(a);
  }

  /**
   * Sets the failed test value.
   *
   * @param a the failing test value
   */
  public void failWith(A a) {
    failure = Optional.of(a);
  }

  /**
   * Returns a string representation.
   *
   * @return the string representation of the result
   */
  @Override
  public String toString() {
    return "Result{" + "successes=" + successes + ", failure=" + failure + '}';
  }

  /**
   * Returns an output string.
   *
   * @return the output string
   */
  public String output() {
    if (successful()) {
      return String.format("SUCCESS%n%d cases tested", successes.size());
    } else {
      return String.format("FAILED%nwith value %s ", failure.get().toString());
    }
  }

}
