package at.jku.ssw.fp.sect11_3.check4j;

import java.util.List;
import java.util.stream.Collectors;
import java.util.stream.IntStream;
import java.util.stream.Stream;

import at.jku.ssw.fp.coll.FList;

/**
 * Class providing shrinkers for various value types.
 *
 * @author Herbert Praehofer
 *
 */
public class Shrinkers {

  /**
   * An shrinker which does not provide values.
   *
   * @param <E> the type of values (for type checking)
   * @return the shrinker not providing a simple shrunk value
   */
  public static <E> Shrinker<E> empty() {
    return (E e) -> Stream.empty();
  }

  /**
   * A shrinker for positive int values.
   */
  public static Shrinker<Integer> posIntShrinker = x -> Stream.iterate(0, y -> y < x, y -> y + 1);

  /**
   * A shrinker for arbitrary int values.
   */
  public static Shrinker<Integer> intShrinker = new Shrinker<Integer>() {
    @Override
    public Stream<Integer> shrink(Integer x) {
      if (x == 0)
        return Stream.empty();
      else {
        int xH = x / 2;
        return Stream.concat(this.shrink(xH), Stream.of(xH));
      }
    }
  };

  /**
   * A shrinker for {@link FList} lists.
   * Lists are shrunk by removing elements.
   *
   * @param <E> the type of list elements
   * @return a shrinker for lists
   */
  public static <E> Shrinker<FList<E>> fListShrinker() {
    return new Shrinker<FList<E>>() {
      @Override
      public Stream<FList<E>> shrink(FList<E> lst) {
        if (lst.isEmpty())
          return Stream.empty();
        else
          return IntStream.range(0, lst.size()).mapToObj(i -> i).flatMap(i -> {
            FList<E> rLst = lst.removeIdx(i);
            return Stream.concat(this.shrink(rLst), Stream.of(rLst));
          }).distinct();
      }
    };
  }

  /**
   * A shrinker for lists of integers
   */
  public static Shrinker<FList<Integer>> intListShrinker = fListShrinker();


  /**
   * A shrinker for {@link Integer} values.
   */
  public static Shrinker<Integer> integerShrinkerRec = new Shrinker<Integer>() {
    @Override
    public Stream<Integer> shrink(Integer x) {
      if (x == 0)
        return Stream.empty();
      else if (x == 1)
        return Stream.of(0);
      else if (x == 2)
        return Stream.of(0, 1);
      else if (x == -2)
        return Stream.of(0, 1, -1, 2);
      else if (x > 0)
        return concat(Stream.of(0, 1, -1), this.shrink(x / 2), Stream.of(x / 2), this.shrink(x - 1), Stream.of(x - 1))
            .distinct();
      else
        return concat(Stream.of(0, 1, -1, 2), this.shrink(-x / 2), this.shrink(x / 2), Stream.of(-x / 2, x / 2),
            this.shrink(x + 1), Stream.of(x + 1)).distinct();
    }
  };

  /**
   * Concatenates multiple streams to a single stream.
   *
   * @param <A> the type of elements of the stream
   * @param streams the streams to concatenate
   * @return the concatenated stream
   */
  @SafeVarargs
  private static <A> Stream<A> concat(Stream<A>... streams) {
    Stream<A> concatenated = Stream.empty();
    for (Stream<A> s : streams) {
      concatenated = Stream.concat(concatenated, s);
    }
    return concatenated;
  }

  /**
   * A shrinker for double values.
   */
  public static Shrinker<Double> doubleShrinker = x -> {
    if (x == 0)
      return Stream.empty();
    else if (x == 1.0)
      return Stream.of(0.0);
    else if (x == -1.0)
      return Stream.of(0.0, 1.0);
    else if (x > 0)
      return Stream.of(0.0, x / 2.0, x - 1.0);
    else
      return Stream.of(0.0, -x, x + 1.0, x / 2.0, x + 1.0);
  };

  /**
   * A shrinker for char values.
   */
  public static Shrinker<Character> charShrinker = c -> {
    if (c == 'a')
      return Stream.empty();
    else if (Character.isUpperCase(c))
      return Stream.of(Character.toLowerCase(c));
    else if (Character.isLetter(c))
      return Stream.of('a');
    else if (Character.isDigit(c) && (c != '0' && c != '1'))
      return Stream.of('0', '1');
    else if (Character.isDigit(c) && (c == '1'))
      return Stream.of('0');
    else
      return Stream.empty();
  };

  /**
   * A shrinker for {@link FList} lists where a given shrinker is used
   * for shrinking the elements.
   * This shrinker shrinks list by removing elements and shrinking elements
   * by the given shrinker.
   *
   * @param <E> the type of list elements
   * @param eShrinker the shrinker for element values
   * @return the shrinker for lists
   */
  public static <E> Shrinker<FList<E>> fListShrinker(Shrinker<E> eShrinker) {
    return new Shrinker<FList<E>>() {

      @Override
      public Stream<FList<E>> shrink(FList<E> lst) {
        if (lst.isEmpty())
          return Stream.empty();
        else
          return Stream.concat(removeElems(lst), shrinkElems(lst)).distinct();
      }

      private Stream<FList<E>> removeElems(FList<E> lst) {
        List<FList<E>> removed = IntStream.range(0, lst.size()).mapToObj(idx -> lst.removeIdx(idx))
            .collect(Collectors.toList());
        return Stream.concat(removed.stream().flatMap(rLst -> shrink(rLst)), removed.stream());
      }

      private Stream<FList<E>> shrinkElems(FList<E> lst) {
        // return Stream.empty();
        if (lst.isEmpty())
          return Stream.empty();
        else
          return IntStream.range(0, lst.size()).mapToObj(i -> i)
              .flatMap(i -> eShrinker.shrink(lst.get(i)).map(x -> lst.replaceIdx(i, x)))
              .peek(l -> System.out.println("shrinkElems " + l)).flatMap(ls -> Stream.concat(shrink(ls), Stream.of(ls)))
              .distinct();
      }

    };

  }

}
